function [cellprops,bgint,trackprops,tracks,inputs] = celltrack(varargin)

% main cell-tracking program, which will track cells and nuclei over time,
% and be able to assign pulses of gene expression to each cell.

% to do this, we'll need:

% - GFP t- and z-series
% - optional nuclei t-series (z- if desired)

% for tracking, need a measure of the cell co-ordinates.  This can be

% - nuclear positions, if we have them
% - otherwise, have to calculate the cell centre from the segmented GFP.

% also, for fast-moving or quite dense cells, the average GFP expression
% could be used as an additional variable to allocate cells to the correct
% track. How reliable is this? Needs to be tested on sparse cells to see if
% the GFP level is stable enough.
% Further criteria could be the estimated volume and footprint area


% Currently, the images should be in the format exported from Volocity as
% multiple Tiff
% These images are read in using the function vti
% A current work in progress is to replace vti with a generic image loader

% Volocity exports the tif images in separate folders for each time frame,
% with the folder name repeated in the filename.  But this might not be the
% case for the Omera imported files, so this step needs to be flexible and
% as foolproof as possible!
% Create the image input section as a separate function to aid integration
% with the file importer and conversion server.

% For standard file import, need the labels:

% Each time point folder is labelled 'T#####', and within this folder the
% files are labelled 'T#####C##Z###'

wb = waitbar(0,'Initialising..');


cellprops = [];
spotprops = [];

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% INPUT PARSING/CHECKING

% use the parameter-value pair input method.

% default values
fpath = cd;
tlabel = 'T';
clabel = 'C';
zlabel = 'Z';
fileext = 'tif'; % this variable will be used to determined what function
                 % function needs to be called to read the files in.
gfpch = 1;
nucch = 2; % set this to zero if nuclei images are not to be used.
nucslice = []; 
usegui = 0;

nucwid = 5; % default nuclear width
cellsiz = 40; % default cell size parameter

zresize = 0;
surfslice = 1;
Debug = 0;
render = 0;

decon = 0;

method3 = [];
method2 = [];

nucmeth = 'default';
nucstdfact = 0.7;

spots = 0;
addgauss = 0;

docomp = 0;

compvals.fg = 0.0045;
compvals.fr = 0.1167;
compvals.ng = 1901;
compvals.nr = 1678;

% default settings in case we just specify spots=1;
spotsiz = 2.8;
spotmethod = 3; % z-filter - is this the best method for size discrimination?
% spotbgrem = 0;
% spotthresh = 2;

% also want to be able to specify starting and ending time points and
% z-slices

t_start = [];
t_end = [];
z_start = [];
z_end = [];

% possibility of downsizing the images, especially for the volume
% measurement where every slice needs to be segmented, should be included.

if nargout>2
    dotracking = 1;
else
    dotracking = 0;
end
trackcutoff = 80;
trackmethod = 4;
trackconf = 0.7;
goodtracklen = 10;
trackintweight = 0;
trackvolweight = 0;

cellmask = 0;
nucmask = 0;

sfactor = 1;
progresume = 0;

bgthr = 2400;

count = 1;

if nargin==1 && iscell(varargin{1})
    varargin = varargin{1};
% else
%     tempin = varargin;
end

while count<=length(varargin)
%     if ii<length(varargin)
        switch lower(varargin{count})
            case 'progresume'
                % resume using partialresults in the event of a crash
                progresume = 1;
                break
            case {'f','fpath'}
                if ischar(varargin{count+1})
                    fpath = varargin{count+1};
                else
                    error('Folder path must be specified as a string')
                end
                count = count + 2;
            case {'e','ext','fileext','filetype'}
                fileext = varargin{count+1};
                count = count + 2;
            case 'agarsettings'
                gfpch = 3;
                nucch = 4;
                sfactor = 2;
                spotsiz = 2.2;
                spots = 1;
                spotmethod = 9;
                decon = 1;
                method3 = 'latest';
                nucmeth = 'agar';
                count = count + 1;
                
            case 'tempsettings'
                % not advisable to use this other than for testing
                % purposes, because it might get changed!
                sfactor = 1;
                spotsiz = 2.2;
                spots = 1;
                spotmethod = [9,20];
                decon = 1;
                method3 = 'vox';
                nucmeth = 'vox';
                gfpch = 1;
                nucch = 2;
                nucwid = 7;
                cellsiz = 45;
                bgthr = 1880;
                
                count = count + 1;
                
            case 'voxagarsettings'
                sfactor = 1;
                spotsiz = 2;
                spots = 1;
                spotmethod = 20;
                decon = 1;
                method3 = 'vox';
                nucmeth = 'vox';
                gfpch = 1;
                nucch = 2;
                nucwid = 7;
                cellsiz = 45;
                bgthr = 1880;
                
                count = count + 1;
            case 'normalsettings'
                gfpch = 3;
                nucch = 4;
                sfactor = 2;
                spotsiz = 2.2;
                spots = 1;
                spotmethod = 9;
                decon = 1;
                method3 = 'latest';
                nucmeth = 'new';
                count = count + 1;
            case 'tlabel'
                tlabel = varargin{count+1};
                count = count + 2;
            case 'clabel'
                clabel = varargin{count+1};
                count = count + 2;
            case 'zlabel'
                zlabel = varargin{count+1};
                count = count + 2;
            case 'gui'
                usegui = varargin{count+1};
                count = count + 2;
            case {'gfp','gfpch'}
                gfpch = varargin{count+1};
                count = count + 2;
            case {'nuc','nucch'}
                nucch = varargin{count+1};
                count = count + 2;
            case {'t_start','tstart'}
                t_start = varargin{count+1};
                count = count + 2;
            case {'t_end','tend'}
                t_end = varargin{count+1};
                count = count + 2;
            case {'z_start','zstart'}
                z_start = varargin{count+1};
                count = count + 2;
            case {'z_end','zend'}
                z_end = varargin{count+1};
                count = count + 2;
            case 'sfactor'
                sfactor = varargin{count+1};
                count = count + 2;
            case 'spots'
                spots = varargin{count+1};
                count = count + 2;
            case {'spotsiz','spotsize'}
                spotsiz = varargin{count+1};
                count = count + 2;
            case 'spotmethod'
                spotmethod = varargin{count+1};
                count = count + 2;
%             case 'spotbgrem'
%                 spotbgrem = varargin{count+1};
%                 count = count + 2;
            case 'nucslice'
                nucslice = varargin{count+1};
                count = count + 2;
            case 'zresize'
                zresize = varargin{count+1};
                count = count + 2;
            case 'surfslice'
                surfslice = varargin{count+1};
                count = count + 2;
            case 'debug'
                Debug = varargin{count+1};
                count = count + 2;
            case 'method3'
                method3 = varargin{count+1};
                count = count + 2;
            case 'method2'
                method2 = varargin{count+1};
                count = count + 2;
            case 'render'
                render = varargin{count+1};
                count = count + 2;
            case 'spotthresh'
                spotthresh = varargin{count+1};
                count = count + 2;
            case 'decon'
                decon = varargin{count+1};
                count = count + 2;
            case 'nucwid'
                nucwid = varargin{count+1};
                count = count + 2;
            case {'cellsiz','cellsize'}
                cellsiz = varargin{count+1};
                count = count + 2;
            case 'nucmeth'
                nucmeth = varargin{count+1};
                count = count + 2;
            case 'nucstdfact'
                nucstdfact = varargin{count+1};
                count = count + 2;
            case 'dotracking'
                dotracking = varargin{count+1};
                count = count + 2;
            case 'trackcutoff'
                trackcutoff = varargin{count+1};
                dotracking = 1;
                count = count + 2;
            case 'trackmethod'
                trackmethod = varargin{count+1};
                dotracking = 1;
                count = count + 2;
            case 'trackconf'
                trackconf = varargin{count+1};
                dotracking = 1;
                count = count + 2;
            case 'goodtracklen'
                goodtracklen = varargin{count+1};
                dotracking = 1;
                count = count + 2;
            case 'trackintweight'
                trackintweight = varargin{count+1};
                dotracking = 1;
                count = count + 2;
            case 'trackvolweight'
                trackvolweight = varargin{count+1};
                dotracking = 1;
                count = count + 2;
            case 'cellmask'
                cellmask = varargin{count+1};
                count = count + 2;
            case 'nucmask'
                nucmask = varargin{count+1};
                count = count + 2;
            case 'bgthr'
                bgthr = varargin{count+1};
                count = count + 2;
            case 'batchfix'
                % put in to allow the batch script generator to work
                % properly
                count = count + 1;
            case 'mag'
                mag = varargin{count+1};
                count = count + 2;
            case 'pixsize'
                pixsize = varargin{count+1};
                count = count + 2;
            case 'addgauss'
                addgauss = 1;
                count = count + 1;
            case 'docomp'
                docomp = varargin{count+1};
                count = count + 2;
            case 'compvals'
                compvals = varargin{count+1};
                count = count + 2;
            otherwise
                error('Unknown input parameter, best to check this!')
        end
%     else
%         % last input, no pair, this has to be interpreted somehow
%         % if the last input is 'gui', equivalent to 'gui',1
%         if strcmpi(varargin{ii},'gui')
%             usegui = 1;
%         else        
%             % just crash for now
%             error('Inputs must be specified as parameter-value pairs')
%         end
%     end
end

if cellmask
    mkdir cellmask
end

if nucmask
    mkdir nucmask
end


inputs = varargin;
if progresume
    % call celltrack recursively?
    load partialresults.mat cellprops bgint folnames ii inputs
    
    % adjust the starting frame to where we left off
    inputs = [inputs,'tstart',ii+1];
    close(wb)
    [appcellprops,appbgint] = celltrack(inputs);
    
    % now append the newly measured properties
    cellprops = [cellprops, appcellprops];
    bgint = [bgint;appbgint];
    tracks = [];
    trackprops = [];
    return
end

if ~spots
    spotprops = NaN;
end

%%%%END OF INPUT CHECKING%%%%%%%%%%%%%%%%%%%%%%

if usegui
    % bring up a graphical use interface to adjust the parameters.
    % do this once we know more about what is required
end

if isempty(t_start)
    t_start = 1;
end
if isempty(t_end)
    t_end = Inf;
end
if isempty(z_start)
    z_start = 1;
end
if isempty(z_end)
    z_end = Inf;
end

if isinf(z_end)
    % need to find the number of z-slices.  The easiest (but not
    % necessarily quickest) way to do this is read in the first image
    testim = voltiffimportbeta('f',fpath,'e',fileext,'t',1,'c',gfpch,...
        'z','all','tlabel',tlabel,'zlabel',zlabel,'clabel',clabel);
    z_end = size(testim,3);
end

if isempty(nucslice)
    if isempty(z_start) || isempty(z_end)
        nucslice = 'all';
    else
        nucslice = z_start:z_end;
    end
end

% find the length of the video
cd(fpath)

switch fileext
    case 'tif'
        folnames = dir([tlabel '*']);
        folnames(~[folnames.isdir]) = []; % only keep folders
        if numel(folnames)==0
            error('No time folders found in specified directory')
        end
        
        numframes = numel(folnames);
        
    % also need case for omera server import, and if we find an ImageJ
    % plugin
end

if t_start<1
    t_start = 1;
end
if t_start>numframes
    t_start = numframes;
end
if t_end<1
    t_end = 1;
end
if t_end>numframes
    t_end = numframes;
end
if z_start<1
    z_start = 1;
end


% want to be able to use multiple spot detection methods for
% comparison.
% make sure that spdet and bgrem are the same size.

% if numel(spotbgrem)==1 && numel(spotmethod)>1
%     spotbgrem = repmat(spotbgrem,size(spotmethod));
% end
% if numel(spotmethod)==1 && numel(spotbgrem)>1
%     spotmethod = repmat(spotmethod,size(spotbgrem));
% end
       
useframes = t_end - t_start + 1;
timedata = [];

for ii = t_start:t_end
    index = ii - t_start + 1;
    
    try
    % we could estimate the time remaining, windows-style
    % if the time is measured in hours, we only need to update the time
    % every couple of minutes
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % TIME REMAINING ESTIMATION
    if index>1
        tempt = toc;
    else
        tempt = [];
    end
    tic
    if length(timedata)>9
        timedata(1) = [];
    end
    timedata = [timedata tempt];
    % could assign to mod_10 of index rather than removing the start and
    % adding on the end
    
    spf = median(timedata);
    timerem = (useframes - index + 1)*spf;
    hoursrem = floor(timerem/3600);
    minsrem = floor((timerem - 3600*hoursrem)/60);
    secsrem = floor(timerem - 3600*hoursrem - 60*minsrem);
    
    if index>1
        if hoursrem>0
            waitbar(index/(useframes+1),wb,sprintf('Frame %d of %d: %d hours %d min remaining',index,useframes,hoursrem,minsrem))
        elseif minsrem>10
            waitbar(index/(useframes+1),wb,sprintf('Frame %d of %d: %d min remaining',index,useframes,minsrem+1))
        else
            waitbar(index/(useframes+1),wb,sprintf('Frame %d of %d: %d min %d sec remaining',index,useframes,minsrem,secsrem))
        end
    else
        waitbar(index/(useframes+1),wb,sprintf('Frame %d of %d',index,useframes))
    end
    
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % CELL SEGMENTATION

    % as a minimum, this script is designed to track cells over time.  If we
    % want to find gene expression spots too, we should do this within the same
    % loop to avoid having to read it in again.
    
    switch fileext
        case 'tif'
            if index==1
                [cellim,tdigs] = voltiffimportbeta('f',fpath,'e',fileext,'t',ii,'c',gfpch,...
                    'z',z_start:z_end,'tlabel',tlabel,'zlabel',zlabel,'clabel',clabel,'tdigs',true);
            else
                cellim = voltiffimportbeta('f',fpath,'e',fileext,'t',ii,'c',gfpch,...
                    'z',z_start:z_end,'tlabel',tlabel,'zlabel',zlabel,'clabel',clabel,'tdigs',tdigs);
            end
    end
    
    
    if nucch~=0
        % we want to use a DNA image to segment the nuclei and improve the
        % cell tracking.  If this is enabled, the nuclear positions should
        % be used as the cell coordinates for tracking, but we still want
        % further measurements from the GFP image, such as gene expression
        % spots and cell area/volume.
        
        % read in the nucleus image
        switch fileext
            case 'tif'
                if index==1
                    [nucim,nuctdigs] = voltiffimportbeta('f',fpath,'e',fileext,'t',ii,'c',nucch,...
                        'z',nucslice,'tlabel',tlabel,'zlabel',zlabel,'clabel',clabel,'tdigs',true);
                else
                    nucim = voltiffimportbeta('f',fpath,'e',fileext,'t',ii,'c',nucch,...
                        'z',nucslice,'tlabel',tlabel,'zlabel',zlabel,'clabel',clabel,'tdigs',nuctdigs);
                end
        end
        
        % Here's where the compensation gets done
        if docomp
            
        end
        
        
        % perform nuclear segmentation first.
        
        % It might be useful to check whether cellim or nucim are empty,
        % which would imply that we've run out of images for the specified
        % channel. This tends to happen when Volocity has crashed during
        % deconvolution and so some of the images are missing.
        
        if isempty(nucim) || isempty(cellim)
            % channel missing, not likely to reappear in the next frame..
            disp('One of the channels has ended early!')
            
            break
        end
        
        if size(nucim,3)>1 && strcmpi(nucmeth(1:3),'def')
            [pos,Lnuc,nucstats] = segnuc4(nucim,nucwid,decon,nucstdfact,1);
            nucbw = Lnuc>0;
            
            for jj = 1:numel(nucstats)
                % we want to measure the mean GFP intensity inside each
                % nucleus.  This should correspond well to the bgint spot
                % property, but it might also be useful for tracking cells
                % from one frame to the next.
                try
                    nucstats(jj).gfpint = admean(cellim(Lnuc==jj)); % we could erode this a bit if we wanted to
                catch ME
                    rethrow(ME)
                end
            end
        elseif strcmpi(nucmeth,'new')
            [pos,Lnuc,nucstats] = segnuc5(nucim,nucwid,decon,nucstdfact,1);
            nucbw = Lnuc>0;
            
            for jj = 1:numel(nucstats)
                % we want to measure the mean GFP intensity inside each
                % nucleus.  This should correspond well to the bgint spot
                % property, but it might also be useful for tracking cells
                % from one frame to the next.
                try
                    nucstats(jj).gfpint = admean(cellim(Lnuc==jj)); % we could erode this a bit if we wanted to
                catch ME
                    rethrow(ME)
                end
            end
        elseif strcmpi(nucmeth,'agar')
            [pos,Lnuc,nucstats] = segnucagar(nucim,nucwid,'auto',nucstdfact,-3);
            nucbw = Lnuc>0;
            
            for jj = 1:numel(nucstats)
                % we want to measure the mean GFP intensity inside each
                % nucleus.  This should correspond well to the bgint spot
                % property, but it might also be useful for tracking cells
                % from one frame to the next.
                try
                    nucstats(jj).gfpint = admean(cellim(Lnuc==jj)); % we could erode this a bit if we wanted to
                catch ME
                    rethrow(ME)
                end
            end
        elseif strcmpi(nucmeth,'simple')
            % simpler nuclei detection for when there is a lot of noise.
        elseif strcmpi(nucmeth,'vox')
            [pos,Lnuc,nucstats] = segnucvox(nucim,nucwid);
            nucbw = Lnuc>0;
            
            for jj = 1:numel(nucstats)
                % we want to measure the mean GFP intensity inside each
                % nucleus.  This should correspond well to the bgint spot
                % property, but it might also be useful for tracking cells
                % from one frame to the next.
                nucstats(jj).gfpint = admean(cellim(Lnuc==jj)); % we could erode this a bit if we wanted to
            end
        else
            if strcmpi(nucmeth,'flat')
                nucim = max(nucim,[],3);
            end
            [pos,Lnuc,nucstats] = segnuc(nucim);
            nucbw = imclearborder(Lnuc>0);
            % Remember that imclearborder will only work as desired in 2D!
            % If we want to do in 3D, we need to pad the top and bottom z
            % with layers of zeros.
            for jj = 1:numel(nucstats)
                try
                    nucstats(jj).gfpint = admean(cellim(Lnuc==jj)); % we could erode this a bit if we wanted to
                catch ME
                    rethrow(ME)
                end
            end
        end
        
        
        %%%%DEBUG%%%%%%%%%%%
        if Debug
%             figure
%             imshow(label2rgb(Lnuc,'jet','w','shuffle'))
        end
        if index==1
            nfnames = fieldnames(nucstats);
        end
    else
        nucbw = [];
        pos = zeros(0,3);
        Lnuc = [];
        if index==1
            nfnames = '';
        end
        % no nuclei images, try to find the cells just from the GFP signal.
        
    end
    
    choodsiz = 8; % specify this here for now
    
    
    % detecting spots requires many z-slices, but measuring the cell shape
    % and cell volume is slowed down massively by this, so probably only a
    % few of the slices required here.
    
    if zresize
        if zresize>1
            nzout = zresize; % value specifies how many slices to rescale to
        else
            nzout = 8; % arbitrary, can change this depending on the default speed required
        end
    else
        nzout = size(cellim,3);
    end
    if sfactor>1 || zresize
        % scale 3D image using resize3
        usecellim = resize3(cellim,[ceil(size(cellim,1)/sfactor),ceil(size(cellim,2)/sfactor),nzout],'super');
    else
        usecellim = cellim;
    end
    
    slicefactor = size(cellim,3)/nzout;
    
    Lcell2 = [];
    
    if sfactor==1
        if strcmpi(method3,'full')
            [tempcellprops,Lcell] = segcell3full(usecellim,...
                pos./repmat([sfactor,sfactor,size(cellim,3)/nzout],[size(pos,1),1]),...
                [cellsiz/sfactor, cellsiz/sfactor, ceil(size(usecellim,3)/2)]);
        elseif strcmpi(method3,'new')
            [tempcellprops,Lcell] = segcell3new(usecellim,pos./repmat([sfactor,sfactor,size(cellim,3)/nzout],[size(pos,1),1]),...
                [],[],[]); % three optional input parameters can be modified by the user.
        elseif strcmpi(method3,'nuc')
            [tempcellprops,Lcell] = segcell3nuc(usecellim,resize3(Lnuc,size(usecellim),'nearest'),0.7);
        elseif strcmpi(method3,'decon')
            [tempcellprops,Lcell,Lcell2] = segcell4(usecellim,max(resize3(Lnuc,size(usecellim),'nearest'),[],3));
            % default size setting for now
        elseif strcmpi(method3,'bgthresh') || strcmpi(method3,'latest')
            [tempcellprops,Lcell,Lcell2] = segcell5(usecellim,max(resize3(Lnuc,size(usecellim),'nearest'),[],3),[],bgthr);
            % default size setting for now
        elseif strcmpi(method3,'vox')
            [tempcellprops,Lcell,Lcell2] = segcellvox(usecellim,max(Lnuc,[],3),cellsiz,bgthr);
        else
            [tempcellprops,Lcell] = segcell3(usecellim,nucbw,choodsiz,method3,method2);
        end
    else
        if strcmpi(method3,'full')
            try
            [tempcellprops,Lcell] = segcell3full(usecellim,...
                pos./repmat([sfactor,sfactor,size(cellim,3)/nzout],[size(pos,1),1]),...
                [40/sfactor, 40/sfactor, ceil(size(usecellim,3)/2)]);
            catch ME
                rethrow(ME)
            end
        elseif strcmpi(method3,'new')
            [tempcellprops,Lcell] = segcell3new(usecellim,pos./repmat([sfactor,sfactor,size(cellim,3)/nzout],[size(pos,1),1]),...
                [],[],[]); % three optional input parameters can be modified by the user.
        elseif strcmpi(method3,'nuc')
            [tempcellprops,Lcell] = segcell3nuc(usecellim,resize3(Lnuc,size(usecellim),'nearest'),0.7);
        elseif strcmpi(method3,'decon')
            [tempcellprops,Lcell] = segcell4(usecellim,max(resize3(Lnuc,size(usecellim),'nearest'),[],3));
            % default size setting for now
        elseif strcmpi(method3,'bgthresh') || strcmpi(method3,'latest')
            [tempcellprops,Lcell,Lcell2] = segcell5(usecellim,max(resize3(Lnuc,size(usecellim),'nearest'),[],3),[],bgthr);
        elseif strcmpi(method3,'vox')
            error('Not tested for scaled images yet')
        else
            if ~isempty(nucbw)
                tempnucbw = logical(imshrink(nucbw,[sfactor,sfactor,1],'max'));
            else
                tempnucbw = nucbw;
            end
            [tempcellprops,Lcell] = segcell3(usecellim,tempnucbw,choodsiz/sfactor,method3,method2);

        end
        Lcell = resize3(Lcell,size(cellim),'nearest');
        if isempty(Lcell2)
            Lcell2 = max(Lcell,[],3);
        end
    
    end
    
    if ~isempty(tempcellprops)
        % once we've got the segmented volumes, use this to measure background
        % intensity in each frame.  We can do this frame by frame, and z-slice
        % by z-slice, but can we try to retain any x-y information?
        bgim = cellim;
        bgim(Lcell>0) = NaN;
        bgint(ii,:) = reshape(admean(admean(bgim,1),2),[1 size(cellim,3) 1]);
        clear bgim
        if ~(strcmpi(method3,'bgthresh') || strcmpi(method3,'latest'))
            Lcell2 = max(Lcell,[],3);
        end

        if ~isempty(pos)
            nxy = round(field2val2(nucstats,'xyz'));
            nuccellind = zeros(size(nxy,1),1);
            for jj = 1:size(nxy,1)
                if ~isnan(nxy(jj,1)+nxy(jj,2))
                    % record the cell index for each nucleus
                    nuccellind(jj) = Lcell2(nxy(jj,1),nxy(jj,2));
                else
                    nuccellind(jj) = NaN;
                end
            end
        end

        if Debug
            size(Lcell)
            length(tempcellprops)
        end
        % here's the point at which to add 2D shape information for cells
        % and nuclei
        try
            bstats = addborder2(Lcell2);
            tempcellprops = mergefields(bstats,tempcellprops);
            
            nbstats = addborder2(max(Lnuc,[],3));
            % nuclear fieldnames need a 'nuc' prefix adding to avoid confusion
            % when combined with cell properties.
            
            % technically this renaming could occur when we directly add the
            % fields to tempcellprops, but leave like this for now..
            [nbstats.nuccumL] = nbstats.cumL;
            [nbstats.nucborder] = nbstats.border;
            nbstats = rmfield(nbstats,{'cumL','border'});
            nucstats = mergefields(nbstats,nucstats);
        catch ME
            rethrow(ME)
        end
        
        
            % convert the properties back to the original size
        tfnames = fieldnames(tempcellprops);
        for jj = 1:numel(tempcellprops)
            if sfactor~=1
                tempcellprops(jj).cellxy = tempcellprops(jj).cellxy * sfactor; % should be accurate enough

                tempcellprops(jj).vol = tempcellprops(jj).vol * sfactor^2 * slicefactor;
                if sum(strcmpi(tfnames,'surfvol'))
                    tempcellprops(jj).surfvol = tempcellprops(jj).surfvol * sfactor^2 * slicefactor;
                end
                % cell footprint area, to see how it scales with volume.
                tempcellprops(jj).footprint = nnz(Lcell(:,:,1)==jj);

                if sum(strcmpi(tfnames,'perim2'))
                    tempcellprops(jj).perim2 = tempcellprops(jj).perim2 * sfactor;
                end
                if sum(strcmpi(tfnames,'area'))
                    tempcellprops(jj).area = tempcellprops(jj).area * sfactor^2;
                end


                if sum(strcmpi(tfnames,'v')) && ~isempty(tempcellprops(jj).d)
                    try
                        if tempcellprops(jj).d(1)>tempcellprops(jj).d(4)
                            tempcellprops(jj).aspr = sqrt(tempcellprops(jj).d(1)/tempcellprops(jj).d(4));
                            tempcellprops(jj).longax = atan(tempcellprops(jj).v(2)/tempcellprops(jj).v(1));
                        else
                            tempcellprops(jj).aspr = sqrt(tempcellprops(jj).d(4)/tempcellprops(jj).d(1));
                            tempcellprops(jj).longax = atan(tempcellprops(jj).v(4)/tempcellprops(jj).v(3));
                        end
                    catch ME
                        rethrow(ME)
                    end
                end

            end
            tempcellprops(jj).frame = ii; % record the actual frame, in case we're stitching together
            if spots
                % initialise each cellprops element as not expressing
                tempcellprops(jj).hasspot = 0;
                if numel(spotmethod)>1
                    tempcellprops(jj).spotmeth = [];
                    % field to specify which methods have detected a spot
                end
            end

            if ~isempty(pos)
                % we have nucleus information to add in
                % firstly, have to assign each nucleus to the correct cell.

                nucind = find(nuccellind==jj);
                if ~isempty(nucind)

                    try
                        tempcellprops(jj).nucxy = nucstats(nucind).xyz;
                    catch ME
                        rethrow(ME)
                    end

                    try
                        tempcellprops(jj).nucarea = nucstats(nucind).Area;
                    catch ME
                        rethrow(ME)
                    end
                    tempcellprops(jj).nucsumint = nucstats(nucind).sumint;
                    % keep as the sum, since this should be invariant with
                    % volume changes eg mitosis
                    if sum(strcmpi(nfnames,'edgepix'))
                        tempcellprops(jj).nucedgepix = nucstats(nucind).edgepix;
                    end
                    tempcellprops(jj).nucbb = nucstats(nucind).BoundingBox;
                    tempcellprops(jj).nucgfp = nucstats(nucind).gfpint;
                    if sum(strcmpi(nfnames,'fluct'))
                        tempcellprops(jj).nucfluct = nucstats(nucind).fluct;
                    end

                    % if we have nucleus coordinates, use these as the default
                    % tracking coordinates
                    tempcellprops(jj).xy = tempcellprops(jj).nucxy;
                else
                    % no nucleus has been found for this cell.  All we can
                    % really do here is highlight it for removal.
                    tempcellprops(jj).xy = tempcellprops(jj).cellxy;
                    tempcellprops(jj).nucsumint = NaN;
                end

            else
                % record the cell coordinates as the default tracking
                % coordinates, since we don't have anything else.

                tempcellprops(jj).xy = tempcellprops(jj).cellxy;
            end
        end
        % need code in segcell3 for measuring the total intensity within a
        % cell. (done, but the background removal is not very good so far)

        if spots

            % for assignment purposes, we don't want to measure spots in nuclei
            % which haven't been assigned to a cell (either because it is on
            % the edge, or some other reason)

            % so modify the nuclear mask to eliminate these

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % SPOT DETECTION

        % If required, also detect gene expression spots while the image is in
        % the memory.


        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Integration of new spot detection algorithm
        % - The labelled (slightly dilated) nuclear image or cell image is used
        %   to find the most spot-like region in the cell.  Using the right
        %   filtering, this should allow spots to be quantified below the
        %   detection threshold
        % - In doing this, we already know which cell each spot belongs to, so
        %   the tracking part of the code below can be taken out.

            % maskim is a binary image used for validating the spots; it should
            % be set to nucbw if we have a nuclei channel, or Lcell>0 otherwise
            if ~isempty(nucbw)
                % modify the nuclear mask in case some of the nuclei haven't
                % been assigned to cells (on the edge, for example)
                % remove these nuclei from the mask, Lcell should cover them
                % all.

                % we're using Lnuc instead of nucbw in the new version of
                % findspots.

                % since the nuclei have already been assigned to cells, it
                % would make sense to relabel the nuclei mask to have the
                % number of the cell it was assigned to.

    %             if size(nucbw,3)==size(Lcell,3)
    %                 maskim = Lcell.*imdilate(nucbw>0,ones(3,3));
    %             else
    %                 maskim = Lcell.*imdilate(repmat(nucbw>0,[1 1 size(Lcell,3)]),ones(3,3));
    %             end
                if size(Lnuc,3)<size(Lcell,3)
                    Lnuc = repmat(max(Lnuc,[],3),[1,1,size(Lcell,3)]);
                end
                maskim = zeros(size(Lnuc));
                for kk = 1:max(Lnuc(:))
                    if ~isnan(nuccellind(kk))
                        maskim(Lnuc==kk) = nuccellind(kk);
                    end
                end

    %             maskim = max(nucbw,[],3) & max(Lcell>0,[],3);

            elseif exist('Lcell','var') && ~isempty(Lcell)
                maskim = repmat(max(Lcell,[],3),[1,1,size(Lcell,3)]);
            else
                maskim = [];
            end

            for kk = 1:numel(spotmethod)

    %             spotoptions.spdet = spotmethod(kk);
    %             spotoptions.bgrem = spotbgrem(kk);
    %             if length(spotthresh)>=kk && ~isnan(spotthresh(kk))
    %                 spotoptions.thresh = spotthresh(kk);
    %             else
    %                 spotoptions.thresh = [];
    %             end
    %             if spotmethod(kk)==6
    %                 spotoptions.cellmask = Lcell;
    %             end

                % if the same background removal method is to be used in more
                % than one case, it is not optimal to run the findspots3
                % algorithm multiple times.  Since the multiple runs are the
                % purposes of finding which spot detection method works best
                % (not for actual processing), leave for now.
                if spotmethod(kk)==20
                    tempspotprops = findspots5(cellim,Lnuc,spotsiz/2*[1,1,1.5],0,'adam');
                else
                    tempspotprops = findspots4(cellim,spotsiz,imdilate(maskim,[0,1,0;1,1,1;0,1,0]),spotmethod(kk),addgauss);
                end

                if numel(spotmethod)>1
                    spotlabel = ['spotprops' num2str(kk)];
                else
                    spotlabel = 'spotprops';
                end
                if numel(tempcellprops) > 0
                    tempcellprops(1).(spotlabel) = [];
                end

                % tempspotprops has been assigned according to the labelling in
                % Lcell, so doesn't need tracking.

                for jj = 1:numel(tempspotprops)
                    tempspotprops(jj).frame = ii;
                    % would also be a good idea to assign the spot to a particular
                    % cell at this stage, while we still have the label matrices in
                    % memory.
                    % If the spot has to lie inside the nucleus, it's almost
                    % certainly going to be the closest position, but for the cell
                    % coordinates only, this isn't so clear cut.

                    nucind = jj;

                    if ~isempty(pos) % nucleus flag
    %                     nucdist = distance(field2val(tempcellprops,'nucxy'),tempspotprops(jj).xy'); %
    %                     nucind = find(nucdist==min(nucdist),1,'first');

                        % can sometimes get NaNs in the tempcellprops.  It
                        % would be best to remove these before we get to the
                        % spot detection stage, but to avoid this causing a
                        % problem, check for NaNs here and at the end.
    %                     try
    %                         ~isempty(tempcellprops(jj).nucxy) && ~isnan(tempcellprops(jj).nucxy(1)) && ~isempty(tempspotprops(jj).xy)
    %                     catch ME
    %                         rethrow(ME)
    %                     end
    try
        ~isempty(tempcellprops(jj).nucxy) && ~isnan(tempcellprops(jj).nucxy(1)) && ~isempty(tempspotprops(jj).xy);
    catch
        keyboard
    end
                        if ~isempty(tempcellprops(jj).nucxy) && ~isnan(tempcellprops(jj).nucxy(1)) && ~isempty(tempspotprops(jj).xy)
                            try
                                nucdist = distance(tempcellprops(jj).nucxy,tempspotprops(jj).xy(:,1:numel(tempcellprops(jj).nucxy)));
                            catch ME
                                rethrow(ME)
                            end
                            tempspotprops(jj).nucdist = nucdist;

                            % also add in the distance to the edge of the nucleus,
                            % where appropriate.  We can get better estimates
                            % of this in deconvolved images; if decon=1, first
                            % remeasure the edge coordinates.

                            nucedgedist = distance(tempspotprops(jj).xy(:,1:size(tempcellprops(nucind).nucedgepix,2)),tempcellprops(nucind).nucedgepix);
                            tempspotprops(jj).nucedgedist = min(nucedgedist);
                        else
                            % need to do anything here?
                        end

                    else

    %                     celldist = distance(tempcellprops(nucind).xy,tempspotprops(jj).xy);
                        tempspotprops(jj).nucdist = [];
                    end
                    % is it allowed for multiple expression spots to be present
                    % in a single cell?
                    if index>1
                        tempspotprops(jj).celllabel = nucind + length(cellprops); % needed only if we're not going to transfer
                                                        % all the info to cellprops
                    else
                        tempspotprops(jj).celllabel = nucind;
                    end

                    % we can transfer the spot properties over to the cell
                    % properties variable, or we could just label the spot
                    % properties so that we know which cell it belongs to.


                    % transfer across - for now, allow for the fact that multiple
                    % expression spots may appear in a cell; may need correcting
                    % once we're happy that the spot detection works OK

                    % how do we keep the spot properties separate for each
                    % method?


    %                 tempcellprops(nucind).spotprops = [tempcellprops(nucind).spotprops;tempspotprops(jj)];
    %                 tempcellprops(nucind).hasspot = 1;

                    tempcellprops(nucind).(spotlabel) = [tempcellprops(nucind).(spotlabel);tempspotprops(jj)];
                    % array growing probably not necessary since we're finding
                    % one and only one spot in every cell

                    tempcellprops(nucind).hasspot = 1;

                    if numel(spotmethod)>1
                        % record which method has detected a spot
                        % this should be obsolete now we're measuring the
                        % spot intensity in every cell in every frame.
                        tempcellprops(nucind).spotmeth = [tempcellprops(nucind).spotmeth, spotmethod(kk)];
                    end

                end

                if numel(spotmethod)>1
                    tspotprops{kk} = tempspotprops;
                else
                    % nothing to do here
                end


            end

        end
    else
        if numel(spotmethod)>1
            tspotprops = cell(1,numel(spotmethod));
        else
            tempspotprops = [];
        end
    end
    
    
    if index~=1
%         cellprops = matchcelltracks(cellprops,tempcellprops,{'xy','vol','int'},100,[0 1 1],[1 110 110]);
        % scale factors currently chosen so that a 20% change in volume or
        % intensity is equivalent to 20 pixels in x or y.
        
%         cellprops = matchcelltracks(cellprops,tempcellprops,{'xy'},100,0,1);
% 
%         % just use the xy for now, until we have more information about how
%         % the cell volume can change, or if it might be better to use
%         % background corrected average intensity.
        
        
        % instead of doing the tracking here, record all the info, and the
        % frame in one long structure vector
        cellprops = [cellprops; tempcellprops(:)]; % still growing inside a loop, is there anyway to avoid this?
        
        if spots
            if numel(spotmethod)>1
                for kk = 1:numel(spotmethod)
                    spotprops{kk} = [spotprops{kk}; tspotprops{kk}(:)];
                end
            else
                try
                    spotprops = [spotprops; tempspotprops(:)];
                catch ME
                    rethrow(ME)
                end
            end
        end
        
        % currently, cells that disappear and reappear become new objects.
    else
        cellprops = tempcellprops(:);
        if spots
            if numel(spotmethod)>1
                spotprops = tspotprops;
            else
                spotprops = tempspotprops(:);
            end
        end
    end
    
    if mod(index,25)==0
        % don't need to save everything, just enough to resume in the event
        % of a crash
        save partialresults.mat cellprops bgint folnames ii inputs
    end
    
    catch ME
        save crashdata
        rethrow(ME)% debug point
    end
    
    if cellmask
        % save the 2D cell label array for later use
        % since we need to know the frame of the movie anyway, we don't
        % need the absolute cellprops number, the number of the cell within
        % each frame will be fine.
        
        Lc = max(Lcell,[],3).*double(sum(Lcell>0,3)>2); %#ok  value can vary to see which looks best
        save(['cellmask/frame' num2str(ii,'%4.4d')],'Lc')
    end
    
    if nucmask
        Ln = max(Lnuc,[],3); %#ok
        save(['nucmask/frame' num2str(ii,'%4.4d')],'Ln')
    end
    
    if render==1
        cellsurf = sum(Lcell>0,3);
        cs2 = imclose(cellsurf,ones(2));
        [x,y,z] = makegrid2(Lcell);
        cs2(imerode(cs2==0,ones(3)))=NaN;
        fvcell = surf2patch(cs2);
        if nucch~=0
            fvnuc = isosurface(y,x,z-2,Lnuc>0,0.5);
        end
        figure
        if nucch~=0
            p2 = patchlight(fvnuc,'red');
            hold on
        end
        p = patchlight(fvcell,'green');
        if nucch~=0
            alpha(0.4)
        end
        psub = patch([0 0 512 512],[0 512 512 0],[0 0 0 0],[0.3 0.6 1]);
        view([-45,30])
        print('-dpng','-r600',['segcells' num2str(ii,['%06.' num2str(tdigs) 'd'])])
        close
    elseif render==2
        % povray export of individual cell meshes
        % save each in a separate directory for each frame
        
        povstr = ['povframe' num2str(ii,'%4.4d')];
        mkdir(povstr)
        cd(povstr)
        
        smoothpovrayexport(Lcell,Lnuc,tempcellprops)
        
        cd ..
        
    end
end

close(wb)


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% TRACKING AND OUTPUT

% get rid of this?

if dotracking
    switch trackmethod
        case 4
            cellprops = performtracking4(cellprops,trackcutoff,trackintweight,trackvolweight,trackconf);
        case 3
            cellprops = performtracking3(cellprops,trackcutoff,trackintweight,trackvolweight);
    end
    tracks = findgoodtracks(cellprops,goodtracklen);
    trackprops = measuretrackprops(cellprops,tracks);
else
    tracks = [];
    trackprops = [];
end




