function matches = dendritictrack(prevr,nextr,cutoff,conf,fovsiz)

% statistics toolbox function for now, but will write my own if I get it
% working OK

% conf is currently a single parameter, which covers defining good nodes
% (by the ratio of current node distance to next), and also choosing which
% configurations to keep as matches.  Need to come up with some way of
% modifying conf specifically for objects which are within cutoff of the
% edge of the image.

% version 4:
% - rather than assigning the best next particles to each prev particle, we
%   look globally for matches above a specific weight.

% version 3: 
% - monte carlo matching algorithm modified to remove duplicate
%   configurations when calculating the weightings.  The issue is that
%   slightly better configurations are being weighted far more than slightly
%   worse due to the frequency with which they are chosen.  Another way of
%   achieving this might be to keep the temperature fairly high during the
%   swapping, so that slightly better configurations are not favoured so
%   much.
% - since MC is much slower anyway, the threshold where we switch from full
%   enumeration to MC sampling is raised.


% version 2 - Weighting temperature now increases close to the edge of the
% field of view, reflecting the decreasing confidence we have about an
% object being visible in the next frame.



if nargin<5 || isemtpy(fovsiz)
    fovsiz = [512 512];
end

if nargin<4 || isempty(conf)
    conf = 0.5;
end
if nargin<3 || isempty(cutoff)
    cutoff = 80;
end

Z = linkage(pdist([prevr;nextr]));

previnds = 1:size(prevr,1);
nextinds = size(prevr,1) + (1:size(nextr,1));
offset = size(prevr,1)+size(nextr,1);
bias = zeros(size(Z,1)+offset,1);
bias(previnds) = 1;
bias(nextinds) = -1;
for ii = 1:size(Z,1)
    bias(ii+offset) = bias(Z(ii,1)) + bias(Z(ii,2));
end

Z(:,4) = bias((offset+1):end);
Zf = [NaN*ones(offset,1),NaN*ones(offset,1),zeros(offset,1),bias(1:offset)];
Zf = [Zf;Z];

for ii = 1:size(Zf,1)
    ind = find(Zf(:,1)==ii | Zf(:,2)==ii);
    if ~isempty(ind)
        Zf(ii,5) = Zf(ind,3);
    else
        Zf(ii,5) = Inf;
    end
end

zz = Zf((size(nextr,1)+size(prevr,1)+1):end,:);
topnode = find(zz(:,3)<=cutoff & (zz(:,5)>cutoff | isinf(zz(:,5)) | isnan(zz(:,5))));
gnode = find(zz(:,3)./zz(:,5) < (conf^1.5) & zz(:,5)<=cutoff);

% Another criteria for defining a good node might be to look for nodes who
% have zero bias and whose parents also have zero bias.  Might not be
% robust enough though..

bad = false(size(gnode));
for ii = 1:numel(gnode)
%     [ii,gnode(ii),zz(gnode(ii),4)]
    % check to see if this can be accepted as a good cluster edge
    % does it have zero bias?
    if zz(gnode(ii),4)~=0
        % if not, is there a higher node within the cutoff which has a zero
        % or opposite bias? If so, reject this gnode
        currnode = gnode(ii);
        nextnode = currnode;
        while sum(nextnode==topnode)==0
            % look at the next node up
            nextnode = find(zz(:,1)==(nextnode+offset) | zz(:,2)==(nextnode+offset));
%             try
%                 abs(zz(nextnode,4))<abs(zz(currnode,4)) || sign(zz(currnode,4))~=sign(zz(nextnode,4));
%             catch ME
%                 rethrow(ME)
%             end
            try
            if abs(zz(nextnode,4))<abs(zz(currnode,4)) || sign(zz(currnode,4))~=sign(zz(nextnode,4))
                % reject this as a good cluster edge
                bad(ii) = true;
                break
            end
            catch ME
                rethrow(ME)
            end
        end
    end
end

% remove those nodes which have been rejected
gnode(bad) = [];

% now, the edges of clusters are given by the nodes in gnode and topnode.
% Need to group the objects based on this information.

gassign = zeros(offset,1);
edgenodes = unique([gnode;topnode]);

% sort the nodes into order of increasing distance
% this should be the same as numerical order, but do this just to make sure
% for now..
[dd,ix] = sort(zz(edgenodes,3),'ascend');

edgenodes = edgenodes(ix);

for ii = 1:numel(edgenodes)
    % assign all the objects within this group
    % work backwards to identify the list of individual objects, and remove
    % any that have already been assigned.
    tempobj = zz(edgenodes(ii),1:2);
    % recursively expand each element of tempobj until we have a list of
    % individual objects
    objlist = getobjlist(tempobj,zz,offset);
    
    objlist(gassign(objlist)~=0) = [];
    gassign(objlist) = ii;
end

% groupings appears to work, we might want to add some more ways of
% defining a good node to cut down the sizes of large groups.
% Failing that, Monte Carlo sampling to minimise the SSD should be fairly
% robust when the number of objects is fixed, ie we fix at the start the
% number of objects which appear or disappear.

matches = zeros(0,2);

for ii = 1:max(gassign)
    gprev = find(gassign==ii & (1:offset)'<=size(prevr,1));
    gnext = find(gassign==ii & (1:offset)'>size(prevr,1)) - size(prevr,1);
    
    if numel(gprev)==1 && numel(gnext)==1
        % simple one-to-one match
        tmatches = [gprev, gnext];
    elseif numel(gnext)==0
        % nothing in the next frame, these objects are lost
        tmatches = zeros(0,2);
    elseif numel(gprev)==0
        % nothing in the previous frame, these objects are new
        tmatches = zeros(0,2);
        
    elseif numel(gprev)>1 || numel(gnext)>1
        % need to look at the SSD to match the objects up
        gprevr = prevr(gprev,:);
        gnextr = nextr(gnext,:);
        
        % decide on the size limits at which we switch to Monte Carlo from
        % full enumeration.
        nobj = max([numel(gprev),numel(gnext)]);
        
        if nobj<8
            ttmatches = fullmatch(gprevr,gnextr,conf,cutoff,fovsiz);
        else
            ttmatches = mcmatch(gprevr,gnextr,conf,cutoff,50*nobj,fovsiz);
        end
        % remove matches with non-existent objects (ie ones which appear or
        % disappear)
        
        
        % as an additional step, we could check for smaller subgroups of
        % 2x2 or 3x3 which haven't been separated and check for further
        % separation.  See if this does anything..
        % find groups first:
        
        
        if ~isempty(ttmatches)
            ttmatches((ttmatches(:,1)>numel(gprev) | ttmatches(:,2)>numel(gnext)),:) = [];
            tmatches = [gprev(ttmatches(:,1)) , gnext(ttmatches(:,2))];
        else
            tmatches = zeros(0,2);
            % flag the occurence of this somehow?
        end
        
    
%     elseif numel(gprev)==1 % for now, include these options in the
%     matching algorthms above
%         % one previous matching multiple nexts, match the closest and any
%         % within 1/f?
%     elseif numel(gnext)==1
        % one next from multiple previous, match the closest and any within
        % 1/f?
    end
    
%     matches = [matches;tmatches+[zeros(size(tmatches(:,1))),size(prevr,1) * ones(size(tmatches(:,1)))]];
    matches = [matches;tmatches];
end


function objlist = getobjlist(tempobj,zz,offset)
% recursive function for obtaining the list of individual objects in the
% group
objlist = [];
for ii = 1:numel(tempobj)
    if tempobj(ii)<=offset
        % individual objects
        objlist = [objlist,tempobj(ii)];
    else
        % find the next pair of nodes down
        nextobj = zz(tempobj(ii)-offset,1:2);
        objlist = [objlist,getobjlist(nextobj,zz,offset)];
    end
end


function matches = mcmatch(prevr,nextr,f,cutoff,n,fovsiz)
% Monte Carlo sampling to find best matches

if nargin<6 || isempty(fovsiz)
    fovsiz = [512 512];
end
if nargin<5 || isempty(n)
    n = 20*size(prevr,1);
end

% first pad the shorter of prev and next with NaNs

% Have to allow somewhere for the fact that a cell disappearing and another
% one appearing could be the best scenario

if size(prevr,1)<size(nextr,1)
    prevr((size(prevr,1)+1):size(nextr,1),:) = NaN;
    uneven = 1;
elseif size(nextr,1)<size(prevr,1)
    nextr((size(nextr,1)+1):size(prevr,1),:) = NaN;
    uneven = 1;
else
    uneven = 0;
end

fakeprevinds = size(prevr,1)+(1:2);
fakenextinds = size(nextr,1)+(1:2);

prevr(fakeprevinds,:) = NaN;
nextr(fakenextinds,:) = NaN;

prevmat = repmat(permute(prevr,[1 3 2]),[1,size(nextr,1),1]);
nextmat = repmat(permute(nextr,[3,1,2]),[size(prevr,1),1,1]);

for ii = 1:size(prevr,1)
    preved(ii,1) = min([prevr(ii,1)-1, fovsiz(1) - prevr(ii,1), ...
        prevr(ii,2)-1, fovsiz(2) - prevr(ii,2)]);
end
for ii = 1:size(nextr,1)
    nexted(ii,1) = min([nextr(ii,1)-1, fovsiz(1) - nextr(ii,1), ...
        nextr(ii,2)-1, fovsiz(2) - nextr(ii,2)]);
end

preved(isnan(preved)) = Inf;
nexted(isnan(nexted)) = Inf;

diffmat2 = sum((nextmat - prevmat).^2,3);
mdist2 = admean(diffmat2(:));

naninds = findn(isnan(diffmat2));
edist = min(preved(naninds(:,1)),nexted(naninds(:,2)));

diffmat2(isnan(diffmat2)) = (0.6+1.2*edist/cutoff)*(cutoff^2); % modify the cost of having an unassigned
% object based on it's distance from the edge

diffmat2(isinf(diffmat2)) = 0; % no penalty for matching two imaginary objects


nobj = size(prevr,1);

% map = zeros(nobj);
configs = zeros(n,size(prevr,1));
ssd = zeros(n,1);

loops = 500; % this can be dependent on the number of objects we have, as can the n in the inputs
% if the aim is not to find the perfect minimum every time, but to explore
% the configurations around the minimum, it might be better to have a small
% number of loops but a larger number of repeats.

for ii = 1:n
    for jj = 1:loops
        T = nobj*mdist2*(loops+1-jj)/loops;
        % a fixed temperature might generate good but non-optimal matches
        % more reliably.
        
        if jj==1
            % generate the first configuration at random
            a0 = randperm(nobj);
            % calculate the SSD in full for the first time
            ssd0 = 0;
            for mm = 1:nobj
                ssd0 = ssd0 + diffmat2(a0(mm),mm);
            end
            overallmin = ssd0;
            minconfig = a0;
        else
            % perform a number of swaps and accept with
            % temperature-dependent probability
%             a1 = a0;
            ssd1 = ssd0;
            % now mutate these by changing a number of indices
            nswaps = 1+floor(rand * (1 + 3*(loops+1-jj)/loops));
            % should be somewhere between 0 and nntnew/2, getting smaller as
            % we get towards the end of the iterations
            tta = a0;
            
            swapinds = ceil(rand(nswaps,2)*nobj);
            for kk = 1:nswaps
                % pick two objects and swap their matches
                ta = tta;
%                 swapind = ceil(rand(2,1)*nobj);
                
                tta(swapinds(kk,1)) = ta(swapinds(kk,2));
                tta(swapinds(kk,2)) = ta(swapinds(kk,1));
                
                % adjust the SSD accordingly
                ssd1 = ssd1 - diffmat2(ta(swapinds(kk,1)),swapinds(kk,1)) - diffmat2(ta(swapinds(kk,2)),swapinds(kk,2)) ...
                    + diffmat2(tta(swapinds(kk,1)),swapinds(kk,1)) + diffmat2(tta(swapinds(kk,2)),swapinds(kk,2));
            end
            a1 = tta;

            % now let's check to see if we keep the result of the swap
            P = exp(-(ssd1 - ssd0)/T);
            if P>rand
                a0 = a1;
                ssd0 = ssd1;
                if ssd0<overallmin
                    overallmin = ssd0;
                    minconfig = a0;
                end
            end
        end
    end
    % update the map
    configs(ii,:) = minconfig;
    ssd(ii) = overallmin;
%     for jj = 1:nobj
%         map(minconfig(jj),jj) = map(minconfig(jj),jj) + 1;
%     end
end

% we now want to remove duplicate configurations and keep only unique ones.
% Hopefully if the temperature hasn't decreased too much and the number of
% loops isn't too high, there'll be a decent distribution here.
% How slow are the built-in functions?

[uconfigs,idx] = unique(configs,'rows');
ussd = ssd(idx);


minssd = min(ussd);
% ginds = find(ssd<(minssd/f));
% use the temperature to find these configurations?


% Weighting by the total SSD ensures good global minimization, but the
% trouble with this is that the final result will have some imperfect
% matches (swapped around) because we've only found local minima at a time.
% Weighting by raw number of configurations has the opposite problem,
% matches are proportional to the distance between prev and next,
% regardless of the effect on global minimization, unless many loops are
% performed.


% matches = zeros(0,2);
totw = zeros(nobj);

for ii = 1:nobj
    % ii represents the previous particle label
    
    % modify confidence based on how close to the edge the previous
    % particle is.
    % assume first two columns of prevr are x and y.
%     if uneven
%         minedgedist = min([prevr(ii,1)-1, fovsiz(1) - prevr(ii,1), ...
%             prevr(ii,2)-1, fovsiz(2) - prevr(ii,2), sqrt(2/size(prevr,2))*cutoff]);
%         % the sqrt factor above is there to account for increases in cutoff due
%         % to volume or intensity tracking.
%         ef = min(1,minedgedist/(sqrt(2/size(prevr,2))*cutoff));
%     else
        ef = 1;
%     end
    
    useT = 2*cutoff^2;
    w = exp((minssd-ussd)/useT); % for speed, the minssd is probably not necessary
    
    for jj = 1:nobj
        totw(jj,ii) = sum(w(uconfigs(:,ii)==jj));
    end
end

totw = totw/sum(totw(:));

% the threshold might have to depend on the temperature
% if nxn array, a threshold of 0.5/n will pick out objects which on average
% fill half the weight or a row or column.
% However, as the temperature increases, the weights should begin to
% equalise, and thus the minimum threshold possible should be 1/(n^2).
% Choose something of a compromise between these two?

wth = 0.35/nobj;

matches = findn(totw>wth);

% [sw,ix] = sort(totw,'descend');
% tmatch = ix(1);
% currw = sw(1);
% count = 2;
% while currw<0.5
%     tmatch = [tmatch; ix(count)];
%     currw = currw + sw(count);
%     count = count+1;
% end
% matches = [matches; [ii*ones(size(tmatch)), tmatch]];
% it would probably be better to make sure that the best configuration
% is recorded too, this way every object will have been assigned.

function blocks = allowedbreakdown(allowed)

% break the allowed matrix down into sub-matrices - what's the best way of
% storing these? As a structure array containing p and n fields?

remprev = true(size(allowed,1),1);
remnext = true(size(allowed,2),1);
% Check that the dimesions of these are right!!

count = 1;

while (sum(remprev) + sum(remnext))>0
    % get the first remaining prev
%     [sum(remprev),sum(remnext)]
    currprev = find(remprev,1,'first');
    currnext = find(allowed(currprev,:));
%     oldnext = [];
    oldprev = [];
    while(numel(currprev)~=numel(oldprev))
    
%         oldnext = currnext;
        oldprev = currprev;
        % now add to currprev any indices in the currnext columns
        for kk = 1:numel(currnext)
            currprev = sort(unique([currprev;find(allowed(:,currnext(kk)))]));
        end
        for kk = 1:numel(currprev)
            currnext = sort(unique([currnext,find(allowed(currprev(kk),:))]));
        end
    end
    blocks(count).p = currprev;
    blocks(count).n = currnext;
    count = count+1;
    remprev(currprev) = 0;
    remnext(currnext) = 0;
end


function matches = guidedmatch(prevr,nextr,allowed,cutoff,fovsiz)

% Look at the possibilities given by allowed, and enumerate (if possible)
% which configuration has the smallest ssd

% don't think we need cutoff or fovsiz for this, since we already know
% which pairings are possible.

% but how do we ensure that we explore every possible combination?  Do we
% need to?
% Can we calculate how mnay combinations there are? As an upper limit,
% prod(sum(allowed,1)) gives a rough idea

roughcombs = prod(sum(allowed,1));

% pick matches at random, blank out each row and column and then repeat
% this has the potential to miss out on some combinations?  But let's
% repeat it lots of times so we should hopefully find the best
% configuration at least once.

% This function will work better if we first break our allowed matrix down
% into separate and independent blocks

ntrials = 2*roughcombs;


function matches = fullmatch(prevr,nextr,f,cutoff,fovsiz)

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% If nprev~=nnext, we need to be careful about our confidence level, f.
% In cases where overlaps are possible, mapping both previous objects onto
% one overlapping next object needs to be more likely than mapping the
% nearest and losing the other.
% The edges are also important in influencing our confidence; when objects
% are close to the edge disappearance is more likely, when they are in the
% centre, merging is more likely.

% There is a difference when the number of objects is fixed; we can have
% more confidence in assigning matches, because every object can be
% uniquely identified. 
% NO! At the edges it can still be quite likely that
% one object disappears and another appears.  Want to make sure that a bad
% chain isn't produced by not allowing one object to disappear and another
% to appear some distance away.

if nargin<5 || isempty(fovsiz)
    fovsiz = [512 512];
end

% first pad the shorter of prev and next with NaNs
if size(prevr,1)<size(nextr,1)
    prevr((size(prevr,1)+1):size(nextr,1),:) = NaN;
    uneven = 1;
elseif size(nextr,1)<size(prevr,1)
    nextr((size(nextr,1)+1):size(prevr,1),:) = NaN;
    uneven = 1;
else
    uneven = 0;
end

% Try allowing for disappearing objects by adding a cost for disappearing
% which is slightly more than the cutoff value

% Add space for non-existent objects in prev and next.  Just a couple to
% begin with to see if it helps

fakeprevinds = size(prevr,1)+(1:2);
fakenextinds = size(nextr,1)+(1:2);

prevr(fakeprevinds,:) = NaN;
nextr(fakenextinds,:) = NaN;

prevmat = repmat(permute(prevr,[1 3 2]),[1,size(nextr,1),1]);
nextmat = repmat(permute(nextr,[3,1,2]),[size(prevr,1),1,1]);
for ii = 1:size(prevr,1)
    preved(ii,1) = min([prevr(ii,1)-1, fovsiz(1) - prevr(ii,1), ...
        prevr(ii,2)-1, fovsiz(2) - prevr(ii,2)]);
end
for ii = 1:size(nextr,1)
    nexted(ii,1) = min([nextr(ii,1)-1, fovsiz(1) - nextr(ii,1), ...
        nextr(ii,2)-1, fovsiz(2) - nextr(ii,2)]);
end

preved(isnan(preved)) = 0;
nexted(isnan(nexted)) = 0;

diffmat2 = sum((nextmat - prevmat).^2,3);

naninds = findn(isnan(diffmat2));
edist = min(preved(naninds(:,1)),nexted(naninds(:,2)));

diffmat2(isnan(diffmat2)) = (0.6+1.2*edist/cutoff)*(cutoff^2); % modify the cost of having an unassigned
% object based on it's distance from the edge

nobj = size(prevr,1);

configs = perms(1:nobj);
ssd = zeros(size(configs,1),1);

for ii = 1:size(configs,1)
    for jj = 1:nobj
        ssd(ii) = ssd(ii) + diffmat2(configs(ii,jj),jj);
    end
end

minssd = min(ssd);
% ginds = find(ssd<(minssd/f));
% use the temperature to find these configurations?

totw = zeros(nobj);

% T = 0.12/f*cutoff^2;
% w = exp((minssd-ssd)/T); % for speed, the minssd is probably not necessary


for ii = 1:nobj
    
    % ii represents the previous particle label
    
    % modify confidence based on how close to the edge the previous
    % particle is.
    % assume first two columns of prevr are x and y.
    if uneven
        minedgedist = min([prevr(ii,1)-1, fovsiz(1) - prevr(ii,1), ...
            prevr(ii,2)-1, fovsiz(2) - prevr(ii,2), sqrt(2/size(prevr,2))*cutoff]);
        % the sqrt factor above is there to account for increases in cutoff due
        % to volume or intensity tracking.
        ef = minedgedist/(sqrt(2/size(prevr,2))*cutoff);
    else
        ef = 1;
    end
    
    useT = 0.12/((0.2 + 0.8*ef)*f)*cutoff^2;
    w = exp((minssd-ssd)/useT); % for speed, the minssd is probably not necessary
    
    
    for jj = 1:nobj
        totw(jj,ii) = sum(w(configs(:,ii)==jj));
    end
end

totw = totw/sum(totw(:));

% the threshold might have to depend on the temperature
% if nxn array, a threshold of 0.5/n will pick out objects which on average
% fill half the weight or a row or column.
% However, as the temperature increases, the weights should begin to
% equalise, and thus the minimum threshold possible should be 1/(n^2).
% Choose something of a compromise between these two?

wth = 0.35/nobj;

matches = findn(totw>wth);

% [sw,ix] = sort(totw,'descend');
% tmatch = ix(1);
% currw = sw(1);
% count = 2;
% while currw<0.5
%     tmatch = [tmatch; ix(count)];
%     currw = currw + sw(count);
%     count = count+1;
% end
% matches = [matches; [ii*ones(size(tmatch)), tmatch]];
% it would probably be better to make sure that the best configuration
% is recorded too, this way every object will have been assigned.
