function [spotprops,bws] = findspots4(im3d,spotsiz,mask,method,addgauss)

% Updated version of spot detection algorithm

% Things that have changed:
% - clean up the options input, the background removal is now connected to
%   the detection method
% - mask is a label matrix for either nuclei or cells; it finds the
%   position with the largest spot response within each region.  If mask is
%   a scalar, there are no regions and this is instead the threshold to use
%   in the whole image.


% Might be possible to get a speed increase when we have a mask image by
% cutting out and filtering only the necessary regions of the cell.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Input checking
if nargin<5 || isempty(addgauss)
    addgauss = 0;
end

if nargin<4 || isempty(method)
    method = 1;
end
if nargin<3 || isempty(mask)
    % default threshold depends on spot detection method
    switch method
        case 1
            mask = 0.5;
    end
end

if nargin<2 || isempty(spotsiz)
    spotsiz = 2.2;
end


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Preliminary calculations

[xdim,ydim,zdim] = size(im3d);



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Step 1 - Background removal

% If we're going to mask the non-cell/nucleus regions of the image, does
% the background need to be subtracted?

% options:
% - median filter with neighbourhood larger than spot size
% - image opening to remove objects with size of spot
% - mean filter (for comparison)
% - Fourier transform

% Keep this in 2D for now

diffim = zeros(size(im3d));
ndim = 3; % by default, change it later in the script if necessary

switch method
    case {1,2}
        % median filter
        nhood = discel(2*spotsiz,1); % is a factor of 2 enough?
        medind = ceil(nnz(nhood)/2);
        for ii = 1:zdim
            medim = ordfilt2(im3d(:,:,ii),medind,nhood);
            diffim(:,:,ii) = im3d(:,:,ii) - medim;
        end
%     case 2
%         % image opening
%         nhood = discel(1.2*spotsiz,1);
%         for ii = 1:zdim
%             diffim(:,:,ii) = im3d(:,:,ii) - imopen(im3d(:,:,ii),nhood);
%         end
%         
%     case 3
%         % mean filter
%         
%         % don't bother with this yet
%     case 4
%         % FFT? Could probably use this for removing shot noise too.
    otherwise
        % do nothing
        diffim = im3d;
end
        

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Step 2 - Identify spots

% options:
% - DoG filter (2D or 3D)
% - small median filter to remove shot noise
% - z-filter to identify clusters of pixels with higher than expected
%   intensity
% - Gaussian (or different) correlation filter

bw = false(size(im3d));

% the name of the output from each of the methods below should be the same,
% fim, so that the mask can be applied afterwards

switch method
    case 1
        % DoG filter
        % this will most likely need the background subtraction first.
        
        fi = dog3split(spotsiz/2,3*spotsiz);
        % keep in 2D for now?
        
%         for ii = 1:zdim
            fim = dog2filt(diffim,fi);
%             bw = fim>options.thresh;
%         end
        
    case 2
        % small or diamond median filter
    case 3
        % z-filter
        zhood = discel(5,1,11) - discel(2,1,11);
        % could replicate zhood in the 3rd dimension, but this might slow
        % it down a lot.
        
        phood = ones(3,3,3);
%         phood(2,2,2) = 0; % seems to work better without the hole in the
%         middle

%         for ii = 1:zdim % loop not necessary
            zim = zfilt(diffim,zhood);
            fim = pprodfilt(zim,phood,1);
            
%             bw = logpim > (nnz(phood)*options.thresh); % 16 -> p~1e-7
%         end

        % an alternative selection method would be to threshold at the
        % individual pixel level, then add up the number of neighbours
        % which are above the threshold.  We can then set a lower AND upper
        % limit for the number of true pixels in, say, a 5x5x5
        % neighbourhood, or just use 2x2x2 or 3x3x3 and no upper limit.
        % Check with a few test images to see what reasonable numbers would
        % be.
        
    case 4
        % Gaussian correlation filter
        template = zeros(ceil(6*spotsiz));
        midp = (size(template,1)+1)/2;
        [x,y] = makegrid2(template);
        template = exp(-((x-midp).^2 + (y-midp).^2)/(spotsiz/2).^2);
        
        fim = zeros(size(diffim));
        
        for ii = 1:zdim
            C = normxcorr2(template,padarray(diffim(:,:,ii),ceil([(4*spotsiz+1) (4*spotsiz+1)]),0));
            % might be necessary to pad the array a little more (will then have to correct
            % the cropping below too)

            % crop and keep only the middle of C, the same size as the original
            % image

            padamt = floor((size(C) - [xdim,ydim])/2);

            fim(:,:,ii) = C(padamt(1):(padamt(1)+xdim-1),padamt(2):(padamt(2)+ydim-1));

%             bw(:,:,ii) = C2(:,:,ii)>options.thresh;
        end


    case 5
        % DoG correlation filter?
        
    case 6
        % method employed by Chubb, Singer et al in first paper
        % Find the n pixels with the highest intensity within the cell (or
        % nucleus), if these pixels form a single region in 3D, it's a
        % spot.  In the original case, n=2.
        
        % for the N=2 case, we don't have the requirement of appearing in
        % more than one z-slice.
        
        % start by specifying the detection parameters here
        N = 6;
        M = 3;
        
        % this will involve breaking image down into individual parts for
        % each cell, requiring that the cell mask is passed as an
        % additional input.
        
        if islogical(mask)
            % need to label each cell first
            mask = bwlabeln(mask);
        end
        if size(mask,3)==1 && size(im3d,3)~=1
            mask = repmat(mask,[1 1 size(im3d,3)]);
        end
        
        fim = false(size(im3d));
        
        for ii = 1:max(mask(:))
            % look in each cell and find the n highest intensity pixels
            % if m of them are together, call it a spot.
            imi = im3d;
            imi(mask~=ii) = NaN;
            cellpx = sort(im3d(mask==ii),'descend');
            
            if numel(cellpx)>(N-1)
            
                th = cellpx(N);

                imspot = imi>=th;
                % find the 'area' of the largest contiguous region
                if nnz(imspot)==2
                    % probably faster to compare coordinates
                    pos = findn(imspot);
                    diffpos = diff(pos,1,1);
                    % maximum of one difference in each direction, sum of two
                    try
                    if max(diffpos)<2 && sum(diffpos)<3
                        fim(imspot) = true;
                    end
                    catch ME
                        rethrow(ME)
                    end
                else
                    lspot = bwlabeln(imspot,18);
                    stats = regionprops(lspot,'area','pixelidxlist');
                    keep = find([stats.Area]>=M);

                    if ~isempty(keep)
                        for kk = 1:length(keep)
                            fim(stats(keep(kk)).PixelIdxList) = true;
                        end
                    end

                end
                % from here, we already know which cell the spot belongs to,
                % but for consistency with the other methods, don't record that
                % here.
                
            end
            
        end
    case 7
        % z filter at the individual pixel level
        
        zhood = discel(spotsiz+3,1,2*ceil(spotsiz) + 5) - discel((spotsiz+3)/2,1,2*ceil(spotsiz) + 5);
        % could replicate zhood in the 3rd dimension, but this might slow
        % it down a lot.
        
        phood = ones(3,3,3);
%         phood(2,2,2) = 0; % seems to work better without the hole in the
%         middle

%         for ii = 1:zdim % loop not necessary
            zim = zfilt(diffim,zhood);
            [logpim,logppix] = pprodfilt(zim,phood,1);
%             bw = logpim > (nnz(phood)*options.thresh); % 16 -> p~1e-7
%         end

        % an alternative selection method would be to threshold at the
        % individual pixel level, then add up the number of neighbours
        % which are above the threshold.  We can then set a lower AND upper
        % limit for the number of true pixels in, say, a 5x5x5
        % neighbourhood, or just use 2x2x2 or 3x3x3 and no upper limit.
        % Check with a few test images to see what reasonable numbers would
        % be.
        
%         bw1 = logppix>options.thresh(1);
        fi = dog3split([spotsiz/2,spotsiz/2,1],[4*spotsiz,4*spotsiz,8]);
        
        fim = dog3filt(-logppix,fi);
        
        % keep only white spots within a certain size range
%         bw = dw>options.thresh(2);

    case 8
        % new design, to make use of the oversampling in the z-direction to
        % cut down on the noise in processing deconvolved images.
        
        % median filter in the z-direction first
        fim1 = ordfiltz(diffim,3);
        fim1a = adordfilt(fim1,discel(8,1,15));
%         fim1a = adordfilt(uint16(fim1),ones(13,13));
        
        fi = dog3split(1,10);
        fim = dog3filt(fim1-fim1a,fi);
        
    case 9
        % only filter the regions which it's necessary to
        % median filter in the z-direction first
        fim1 = ordfiltz(diffim,3);
        siz = size(fim1);
%         fim1a = NaN*zeros(siz);
        
        stats = regionprops(mask,'BoundingBox');
        fi = dog3split(1,10);
        
        Lmax = zeros(size(im3d));
        
        for ii = 1:numel(stats)
            xmin = max(1,stats(ii).BoundingBox(1)-6.5);
            xmax = min(siz(1),stats(ii).BoundingBox(1)+stats(ii).BoundingBox(4)+6.5);
            ymin = max(1,stats(ii).BoundingBox(2)-6.5);
            ymax = min(siz(2),stats(ii).BoundingBox(2)+stats(ii).BoundingBox(5)+6.5);
            zmin = max(1,stats(ii).BoundingBox(3)-0.5);
            zmax = min(siz(3),stats(ii).BoundingBox(3)+stats(ii).BoundingBox(6)+0.5);
            
            tempfim = adordfilt(fim1(ymin:ymax,xmin:xmax,zmin:zmax),ones(13,13));
            % x and y swapped due to the way BoundingBox is defined by
            % regionprops
            tdog = dog3filt(fim1(ymin:ymax,xmin:xmax,zmin:zmax)-tempfim,fi);
            
            fim(ymin:ymax,xmin:xmax,zmin:zmax) = tdog;
            
            % also generate bws here
            tmask = mask(ymin:ymax,xmin:xmax,zmin:zmax)==ii;
            tbws = false(size(tmask));
            vim = tdog(tmask);
            maxv = max(vim);
            
            logmax = vim==maxv;
            
            if nnz(logmax)>1
                tim = diffim(ymin:ymax,xmin:xmax,zmin:zmax);
                vtim = tim(tmask);
                vmaxes = vtim(logmax);
                ind = find(vmaxes==max(vmaxes),1,'first');
                lind = find(logmax);
                logmax = zeros(size(logmax));
                logmax(lind(ind)) = true;
            end
            
            tbws(tmask) = logmax;
            Lmax(ymin:ymax,xmin:xmax,zmin:zmax) = max(Lmax(ymin:ymax,xmin:xmax,zmin:zmax),ii*tbws);
            % max required in case of overlapping bounding boxes between
            % adjacent nuclei..
            
        end
        bws = Lmax>0;
    case 10
        % just like 9 but treats each slice individually, for the case
        % where our slices are further apart and so the spots might not
        % appear in multiple slices, or the case where we only have one
        % slice, which also has to work using the same code
        
        if size(diffim,3)>1
            ndim = 3;
            xi = 1;
            xw = 4;
            yi = 2;
            yw = 5;
        else
            ndim = 2;
            xi = 1;
            xw = 3;
            yi = 2;
            yw = 4;
        end
        
        siz = size(diffim);
%         fim1a = NaN*zeros(siz);
        
        stats = regionprops(mask,'BoundingBox');
        fi = dog3split(1,10);
        
        Lmax = zeros(size(im3d));
        
        for ii = 1:numel(stats)
            xmin = max(1,stats(ii).BoundingBox(xi)-6.5);
            xmax = min(siz(1),stats(ii).BoundingBox(xi)+stats(ii).BoundingBox(xw)+6.5);
            ymin = max(1,stats(ii).BoundingBox(yi)-6.5);
            ymax = min(siz(2),stats(ii).BoundingBox(yi)+stats(ii).BoundingBox(yw)+6.5);
            if ndim==3
                zmin = max(1,stats(ii).BoundingBox(3)-0.5);
                zmax = min(siz(3),stats(ii).BoundingBox(3)+stats(ii).BoundingBox(6)+0.5);
            else
                zmin = 1;
                zmax = 1;
            end
            
            tempfim = adordfilt(diffim(ymin:ymax,xmin:xmax,zmin:zmax),ones(13,13));
            % x and y swapped due to the way BoundingBox is defined by
            % regionprops
            tdog = dog2filt(diffim(ymin:ymax,xmin:xmax,zmin:zmax)-tempfim,fi);
            
            fim(ymin:ymax,xmin:xmax,zmin:zmax) = tdog;
            
            % also generate bws here
            tmask = mask(ymin:ymax,xmin:xmax,zmin:zmax)==ii;
            tbws = false(size(tmask));
            vim = tdog(tmask);
            maxv = max(vim);
            
            logmax = vim==maxv;
            
            if nnz(logmax)>1
                tim = diffim(ymin:ymax,xmin:xmax,zmin:zmax);
                vtim = tim(tmask);
                vmaxes = vtim(logmax);
                ind = find(vmaxes==max(vmaxes),1,'first');
                lind = find(logmax);
                logmax = zeros(size(logmax));
                logmax(lind(ind)) = true;
            end
            
            tbws(tmask) = logmax;
            Lmax(ymin:ymax,xmin:xmax,zmin:zmax) = max(Lmax(ymin:ymax,xmin:xmax,zmin:zmax),ii*tbws);
            % max required in case of overlapping bounding boxes between
            % adjacent nuclei..
            
        end
        bws = Lmax>0;
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Step 3 - Check validity

% if we've been given a mask, find the point with the higest response
% within each region, if instead it's a scalar, simply threshold.

% bw = zeros(size(im3d));


if ~isscalar(mask)
    % if the mask is logical, first label with integers.
    if islogical(mask)
        mask = bwlabeln(mask>0);
    end
    ncells = max(mask(:));
    
    if size(mask,3)<size(im3d,3)
        mask = repmat(mask,[1 1 size(im3d,3)]);
    end
    
    % now we need to go through each region and pick out the maximum
    % response to the spot detection filtering, but only if we didn't use
    % method 6, because this method does this already.
    
    % actually, method 6 needs this as well, because we find the area here
    % as well as spot position.
    
    if method~=6 && method~=9 && method~=10
        bws = zeros(size(im3d));
        for ii = 1:ncells
            % first find the pixel with the maximum response within the
            % region.
            tempfim = fim;
            try
                tempfim(mask~=ii) = -Inf;
            catch
                keyboard
            end
            tempmax = tempfim==max(tempfim(:));
            
            if nnz(tempmax)>1
                % take the one with the highest image intensity as well
                tempim = diffim;
                tempim(~tempmax) = -Inf;
                tempmax = tempim==max(tempim(:));
                if nnz(tempmax)>1
                    % still more than one pixel with both maxima identical
                    % just pick the first one..
                    maxinds = findn(tempmax);
                    for kk = 2:size(tempmax,1)
                        tempmax(maxinds(kk,1),maxinds(kk,2),maxinds(kk,3)) = 0;
                    end
                end
            end
            bws = bws + tempmax;
        end
        Lmax = mask.*bws;
    end
    
    
% % %     Lspot = mask.*(bw>0); % simple way of correct spot labelling
    
    
else
% % %     Lspot = bwlabeln(fim>mask); % default connectivity is 26 neighbours

end


% Now get the properties for each spot
% Like the FRAP analysis, we now want to measure the spot intensity
% relative to background rings of increasing radii.

% [x,y,z] = makegrid2(im3d);

% % % tempprops = regionprops(Lspot,'Centroid','PixelIdxList','PixelList','Area');

% % % if method~=1
% % %     % we need a new fim, but don't recalculate it if we already have it.
% % %     fi = dog3split(spotsiz/2,3*spotsiz);
% % %     fim = dog2filt(diffim,fi);
% % % end

pim3d = padarray(im3d,[6,6,1],NaN);

for ii = 1:ncells
    flag = 0;
    if ~isscalar(mask)
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % SLOW LINE!
        tempxyz = findn(bws & mask==ii);
        if ~isempty(tempxyz)
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
            % find the spot pixel
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            % ANOTHER SLOW LINE, MUST BE A QUICKER WAY
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
            try
                tempxyz = tempxyz(1,:);
            catch ME
                rethrow(ME)
            end
            spotprops(ii,1).xyfirst = tempxyz;
            % now find the maximum filter response close to the pixel
% % %             fimreg = getimreg(fim,tempxyz,[3 3 3]);
% % %             maxinds = findn(fimreg==max(fimreg(:)));

            if ndim==3
                spotprops(ii,1).filtresp = fim(tempxyz(1),tempxyz(2),tempxyz(3));
            else
                spotprops(ii,1).filtresp = fim(tempxyz(1),tempxyz(2));
            end

            % if more than one max, take the one closest to the centre
% % %             if size(maxinds,1)>1
% % %                 centdist = sum(abs(2 - maxinds),2);
% % %                 keep = find(centdist==min(centdist,1,'first'));
% % %                 maxinds = maxinds(keep,:);
% % %             end
% % % 
% % %             try
% % %                 tempxyz = tempxyz - [2,2,2] + maxinds;
% % %             catch ME
% % %                 rethrow(ME)
% % %             end

            % now find the peak of fim to subpixel accuracy,
            % if the peak doesn't fall within one pixel, our spot finding
            % algorithm isn't very good.
            fimreg = getimreg(fim,tempxyz,[3 3 3]);
            delta = regsubpix(fimreg,1); % use method 1 for now, since we know it works
            % if absolute value of delta is greater than 0.5, our peak isn't
            % well defined.  Just use it for now
            delta(abs(delta)>1) = 0;
            try
                spotprops(ii,1).xy = tempxyz + delta(1:numel(tempxyz));
            catch ME
                rethrow(ME)
            end
        else
            % not a genuine object, doesn't have a mask.  Probably this
            % object was a nucleus that didn't pick up a cell
            flag = 1;
        end
       
    else
        % with appropriate modification of bws, we could easily use the
        % above method in the case of no cell or nuclear mask.
        tempfim = fim;
        tempfim(Lspot~=ii) = -Inf;
        spotprops(ii,1).xy = findn(tempfim==max(tempfim(:)));
        if size(spotprops(ii,1).xy,1)>1
            spotprops(ii,1).xy = spotprops(ii,1).xy(1,:);
            % maybe a better way of picking here is to choose the one
            % closest to the centroid?
        end
% % %         spotprops(ii,1).xy = max(0.5,tempprops(ii).Centroid([2 1 3]));
        
    end
    
    if ~flag

        if spotprops(ii,1).xy(1)>xdim
            spotprops(ii,1).xy(1)=xdim;
        end
        if spotprops(ii,1).xy(2)>ydim
            spotprops(ii,1).xy(2)=ydim;
        end
        if ndim>2 && spotprops(ii,1).xy(3)>zdim
            spotprops(ii,1).xy(3)=zdim;
        end

%         spotprops(ii,1).area = tempprops(ii).Area;

    %     spotprops(ii,1).spotint = sum(im3d(tempprops(ii).PixelIdxList));
        % for the proper spotint, we should subtract off the background,
        % estimated by the mean intensity of pixels surrounding the spot.
        
        imreg = getimreg(im3d,spotprops(ii,1).xy,[3 3 3]);
        spotprops(ii,1).maxspotint = max(imreg(:));


        % estimate background intensity using a hollow cylinder, outer radius
        % 6, inner radius 3, height 3

    %     spotprops(ii,1).bgint1 = mean(im3d(((x-spotprops(ii).xy(1)).^2 + (y-spotprops(ii).xy(2)).^2 <=25) & ...
    %         ((x-spotprops(ii).xy(1)).^2 + (y-spotprops(ii).xy(2)).^2 >=4) & ...
    %          (abs(z - spotprops(ii).xy(3)) < 1.5)));

        % if the padding operation is too slow, an alternative is to crop the
        % useinds array to avoid going out of the available range.

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % if we only have 3 z-slices, chances are quite high that the spot will
        % be in one of the outer slices, which means that the measured spot
        % intensity will vary systematically depending on if it's in an outer
        % layer or not.  Therefore, in the case where there are only 3 slices
        % (or 5 or less?), sum only in the layer of interest for consistency,
        % rather than the three layers around it.

        if size(im3d,3)<6
            zuse = 1;

            bginds = discel(6,1,13) - discel(3,1,13);
            spotinds = discel(3,1,13);

        else
            zuse = 0:2;

            bginds = repmat(discel(6,1,13) - discel(3,1,13),[1,1,3]);
            spotinds = repmat(discel(3,1,13),[1,1,3]);

        end

        try
            if ndim==3
                spotnhood = pim3d((round(spotprops(ii,1).xy(1)) + (0:12)),(round(spotprops(ii,1).xy(2)) + (0:12)),(round(spotprops(ii,1).xy(3)) + zuse));
            else
                spotnhood = pim3d((round(spotprops(ii,1).xy(1)) + (0:12)),(round(spotprops(ii,1).xy(2)) + (0:12)));
            end
        catch ME
            rethrow(ME)
        end

        if ~isscalar(mask)
            
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            spotprops(ii,1).longbgim = admean(im3d(mask==ii & ~imdilate(Lmax==ii,discel(6.5,1))));
            % ABOVE IS THE SLOWEST LINE OF THE ENTIRE PROGRAM - DO WE
            % ACTUALLY NEED IT?
            %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        else
            spotprops(ii,1).longbgim = NaN;
        end

        try
            spotprops(ii,1).bgint = admean(spotnhood(logical(bginds)));
            spotprops(ii,1).spotint = admean(spotnhood(logical(spotinds)));
            
            % the corrected sum needs to take account of if some of the
            % spot indices lie outside the image and shouldn't contribute.
            spotprops(ii,1).corrsum = adsum(spotnhood(logical(spotinds))) - (spotprops(ii).bgint * nnz(~isnan(spotnhood(logical(spotinds)))));
        catch ME
            rethrow(ME)
        end
        for jj = 1:size(spotnhood,3)
            zraw(jj) = adsum(logref(imreg(:,:,jj),spotinds2d));
            zbg(jj) = admean(logref(imreg(:,:,jj),bginds2d));
            zmedbg(jj) = admedian(logref(imreg(:,:,jj),bginds2d));
            
            npix2d(jj) = nnz(spotinds2d & ~isnan(imreg(:,:,jj)));
            
        end
        spotprops(ii).zraw = zraw;
        spotprops(ii).zbg = zbg;
        spotprops(ii).zspotint = zraw - npix2d.*zbg;
        
        
    else
        % do we have to put anything here in the case where the mask isn't
        % correct?
        % NaNs? or will empty fields be good enough..?
    end
    
end

if addgauss
    xyz = field2val2(spotprops,'xyfirst');
    g2props = gaussspotdata(im3d,xyz,spotsiz/2);
    g3props = gauss3spotdata(im3d,xyz,spotsiz/2);
    
    spotprops = mergefields(g3props,mergefields(g2props,spotprops));
end

if ncells==0
    spotprops = [];
end

