function [spotprops,sxyz] = findspots5(im3d,Lnuc,firad,justxyz,options,preweight)

% new spot finding and measuring function, using experimental methods of
% noise removal and background estimation.

defopt = struct('upfact',2,'downfact',0.5,'uprespfact',3,'downrespfact',0.8,...
    'uppasswidfact',0.15,'downpasswidfact',0.15,'dosubpix',1);

if nargin<6
    preweight = [];
end
if nargin<5 || isempty(options)
    options = defopt;
elseif isstruct(options)
    options = mergefields(defopt,options,0);
else
    switch lower(options)
        case 'adam'
            options = struct('upfact',2,'downfact',0.5,'uprespfact',1/0.3105,...
                'downrespfact',0.6,'uppasswidfact',0.15,'downpasswidfact',0.25);
        case 'broad'
            % slightly broader windows for the ratios
            options = struct('upfact',2,'downfact',0.5,'uprespfact',1/0.3105,...
                'downrespfact',0.6,'uppasswidfact',0.25,'downpasswidfact',0.35);
        otherwise
            options = defopt;
    end
    
    % merge in the default options so that we don't have to specify
    % everything in the str presets above
    options = mergefields(defopt,options,0);
end

if nargin<4 || isempty(justxyz)
    justxyz = 0;
end

if nargin<3 || isempty(firad)
    firad = [1,1,1.5];
end

if isscalar(firad)
    firad = [1,1,1.5]*firad;
end

if size(Lnuc,1)==size(im3d,1) && size(Lnuc,2)==size(im3d,2)
    if size(im3d,3)>1 && size(Lnuc,3)==1
        Lnuc = repmat(Lnuc,[1,1,size(im3d,3)]);
    end
    havecoords = 0;
else
    % we might have supplied the spot coordinates rather than a mask
    havecoords = 1;
    % can't save any filtering time since we record all the response values
    % for later..
    % unless we want to implement a function which automatically cuts out
    % regions around each coordinate, performs the filtering and puts it
    % back into the image (or returns the value at the point of interest,
    % in which case we'd have to flag up that this has been done..)
end

% remove large objects by subtracting off a morphologically opened image.
% Do in 2D only to save time.  This might affect the apparent spot size,
% since large lengthscale components of the spot will be removed, something
% to bear in mind when choosing the filter ratios etc.

opim = imopen(im3d,newdiscel(2.5*firad(1),1,odd(5*firad(1))));
subim = im3d - opim;

% a lot of time can be saved by only filtering the regions where there is a
% nucleus, but we have to be careful not to introduce edge effects.  This
% might be useful enough to have a separate function, if the overhead isn't
% too big.
% On the other hand, the fastest option is probably to find the maximum
% value as we go along..

fi = dog3split2(firad,odd(15*firad));
fi2 = dog3split2(firad*options.upfact,odd(15*firad*options.upfact));
fi3 = dog3split2(firad*options.downfact,odd(15*firad*options.downfact));

sfim = dog3filt(subim,fi);
sfim2 = dog3filt(subim,fi2);
sfim3 = dog3filt(subim,fi3);

epsilon = 0.004*admean(sfim(sfim>0));

sfim(sfim<epsilon) = epsilon;
sfim2(sfim2<epsilon) = epsilon;
sfim3(sfim3<epsilon) = epsilon;

% an alternative to a Gaussian bandpass we could use two sigmoids instead
resp = exp(-(sfim./sfim2 - options.uprespfact).^2 / (2*((options.uprespfact*options.uppasswidfact)^2)));
resp2 = exp(-(sfim./sfim3 - options.downrespfact).^2 / (2*((options.downrespfact*options.downpasswidfact)^2)));

% worth blurring this at all before multiplying sfim?

% In a number of cases, it might be useful to pre-weight our search  based
% on a user-supplied array.  The most pressing use for this is the FRAP
% movies, where the spot disappears, but we know roughly where it should be
% because of bleaching of the H2B signal.

% the ratio can be problematic when negative numbers or numbers close to
% zero are involved, so maybe instead we should just look for the value of
% (sfim - uprespfact*sfim2) and not need sfim in fim?

% another option would be to further weight the pixels based on whether or
% not they are local maxima in sfim, which would reduce the halo effect
% around spots which are a bit too small

lmax = imregionalmax(sfim,18);

if isempty(preweight)
    fim = sfim.*resp.*resp2.*(0.6 + 0.4*double(lmax));
else
    fim = sfim.*resp.*resp2.*(0.6 + 0.4*double(lmax)).*preweight;
end

if ~havecoords
    % then find the maximum value of fim within each nucleus.
    sxyz = NaN*zeros(max(Lnuc(:)),3);
    for ii = 1:max(Lnuc(:))
        temp = fim;
        temp(Lnuc~=ii) = 0;
        temp2 = findn(temp==max(temp(:)));
        if ~isempty(temp2)
            sxyz(ii,:) = temp2(1,:);
        end
    end
else
    sxyz = Lnuc;
end

if justxyz
    spotprops = sxyz;
    return
end

% I wonder if spinds needs to be explicitly recorded as an integer,
% particularly when the image size is extremely large?
spinds = adsub2ind(size(im3d),round(sxyz));

% Then do the spot intensity calculations on each location.
% the image neighbourhood size is currently set here regardless of the spot
% size.

% the nhsiz should really depend on what the spot size is
% make sure that a spot rad of 1 comes out with these settings
% nhsiz = [17,17,7];
% spotrad = 3.5;
% bgrad = 6;
% spoth = 7;

nhsiz = odd([16*firad(1),16*firad(2),4*firad(3)],'up');
spotrad = 3.5*firad(1);
bgrad = 6*firad(1);
spoth = nhsiz(3);

zz = (1:spoth)-0.5*(spoth+1);
zwt = ones(size(zz)); % not used for now..

bginds2d = logical(newdiscel(bgrad,1,nhsiz(1)) - newdiscel(spotrad,1,nhsiz(1)));
spotinds2d = logical(newdiscel(spotrad,1,nhsiz(1)));

bginds = logical(padarray(repmat(bginds2d,[1,1,spoth]),[0,0,(nhsiz(3)-spoth)/2],0,'both'));
spotinds = logical(padarray(repmat(spotinds2d,[1,1,spoth]),[0,0,(nhsiz(3)-spoth)/2],0,'both'));

% % oldbginds = logical(repmat(discel(6,1,13) - discel(3,1,13),[1,1,3]));
% % oldspotinds = logical(repmat(discel(3,1,13),[1,1,3]));

wwid = 6;
xdata = adind2sub(nhsiz,1:prod(nhsiz));
wdata = exp(-((xdata(:,1)-(nhsiz(1)+1)/2).^2 + (xdata(:,2)-(nhsiz(1)+1)/2).^2)/(2*wwid^2));

mask = gauss2filt(spotinds,0.4,0);

for ii = 1:size(sxyz,1)
    if isnan(sxyz(ii,1))
        % set the intensity to zero and everything else to NaN
        spotprops(ii,1).xy = [NaN,NaN,NaN];
        % to make sure that spotprops is the right size..
        
        continue
    end
    if options.dosubpix
        fimreg = getimreg(sfim,sxyz(ii,:),[3,3,3]);
        delta = pivsubpix3(fimreg);
    else
        delta = NaN;
    end
    spotprops(ii,1).xyfirst = sxyz(ii,:);
    if any(~isfinite(delta))
        spotprops(ii).xy = sxyz(ii,:);
    else
        spotprops(ii).xy = sxyz(ii,:) + delta;
    end
    
    % cut out the region around the spot ready to do various measurements..
    imreg = getimreg(im3d,sxyz(ii,:),nhsiz);
    medbg = admedian(imreg(bginds));
    bgint = admean(imreg(bginds));
    spotint = adsum(imreg(spotinds) - bgint);
    rawint = adsum(imreg(spotinds));
    
    % Many of the spot measurement methods do not work very well at all.
    % So some of them are commented out, ahead of being removed in a later
    % version.
    
    spotprops(ii).medbg = medbg;
    spotprops(ii).rawint = rawint;
    spotprops(ii).bgint = bgint;
    spotprops(ii).corrsum = spotint;
    
    % also record the various filter responses
    spotprops(ii).fresp = sfim(spinds(ii));
    spotprops(ii).fresp2 = sfim2(spinds(ii));
    spotprops(ii).fresp3 = sfim3(spinds(ii));
    
    spotprops(ii).totresp = fim(spinds(ii));
    
    % try the thermal equilibrium background estimation
    temp3 = adblurnans2(imreg,mask,0.5,[],1);
    
    fillbg = admean(temp3(spotinds));
    
    fillreg = imreg - temp3;
    fillint = adsum(fillreg(:));
    zfillint = adsum(reshape(fillreg,[size(imreg,1)*size(imreg,2),size(imreg,3)]),1)';
    
%     spotprops(ii).fillrawint = rawint;
    spotprops(ii).fillbg = fillbg;
    spotprops(ii).fillint = fillint;
    spotprops(ii).zfillint = zfillint;
    
    % also do on the open subtracted image to see if this makes a difference
    
    subreg = getimreg(subim,sxyz(ii,:),nhsiz);
    medbg = admedian(subreg(bginds));
    bgint = admean(subreg(bginds));
    spotint = adsum(subreg(spotinds) - bgint);
    rawint = adsum(subreg(spotinds));
    
    spotprops(ii).submedbg = medbg;
    spotprops(ii).subrawint = rawint;
    spotprops(ii).subbgint = bgint;
    spotprops(ii).subcorrsum = spotint;
    
    % possibly a median estimation of the background intensity would be
    % more accurate than the mean, which might be affected by noise.
    
    
    % try the thermal equilibrium background estimation
    temp3 = adblurnans2(subreg,mask,0.5,[],1);
    
    fillbg = admean(temp3(spotinds));
    fillint = adsum(subreg(:)-temp3(:));
    
    zfillint = adsum(reshape(subreg-temp3,[size(subreg,1)*size(subreg,2),size(subreg,3)]),1)';
    
    spotprops(ii).subfillbg = fillbg;
    spotprops(ii).subfillint = fillint;
    spotprops(ii).zsubfillint = zfillint;
    
    % zsub3 is one of the most accurate spot detection methods, so it would
    % make sense to directly implement it here so that it can be referenced
    
    
    % for completeness, also include the old spot detection, which almost
    % certainly misses some intensity from the top and bottom..
    
% %     spotnhood = getimreg(im3d,sxyz(ii,:),[13,13,3]);
% %     spotprops(ii,1).oldbgint = admean(spotnhood(oldbginds));
% %     spotprops(ii,1).oldspotint = admean(spotnhood(oldspotinds));
% %     
% %     % the corrected sum needs to take account of if some of the
% %     % spot indices lie outside the image and shouldn't contribute.
% %     spotprops(ii,1).oldcorrsum = adsum(spotnhood(oldspotinds)) - (spotprops(ii).oldbgint * nnz(~isnan(spotnhood(oldspotinds))));
    
    
    % calculate separate bg and spot values for each z slice, which we can
    % combine however we want later..
    for jj = 1:nhsiz(3)
        zraw(jj) = adsum(logref(imreg(:,:,jj),spotinds2d));
        zbg(jj) = admean(logref(imreg(:,:,jj),bginds2d));
        zmedbg(jj) = admedian(logref(imreg(:,:,jj),bginds2d));
        
        npix2d(jj) = nnz(spotinds2d & ~isnan(imreg(:,:,jj)));
        
    end
    spotprops(ii).zraw = zraw;
    spotprops(ii).zbg = zbg;
    spotprops(ii).zspotint = zraw - npix2d.*zbg;
    spotprops(ii).zspotintmed = zraw - npix2d.*zmedbg;
    
    
    % trial the Gaussian layer fitting?
%     keyboard
% %     try
% %         [temp,r2,db] = gausswfitlayers(imreg(:),xdata,[],wdata);
% %         % is it worth storing the fit values?
% %         % try it for now..
% %         spotprops(ii).glfitdata = temp;
% %         spotprops(ii).glr2 = r2;
% %         spotprops(ii).glsum = sum(flatten(db.withoutbg));
% %         spotprops(ii).glbound = any(db.isbound);
% %     catch ME
% %         % for now if this generates an error, NaN things out and continue.
% %         spotprops(ii).glfitdata = NaN;
% %         spotprops(ii).glr2 = -1;
% %         spotprops(ii).glsum = NaN;
% %         spotprops(ii).glbound = NaN;
% %     end
% %     
% %     
% %     % Also do layered Gaussian fitting on the filled region and
% %     % open-subtracted region for comparison.
% %     try
% %         [temp,r2,db] = gausswfitlayers(fillreg(:),xdata,[],wdata);
% %         spotprops(ii).fillglfitdata = temp;
% %         spotprops(ii).fillglr2 = r2;
% %         spotprops(ii).fillglsum = sum(flatten(db.withoutbg));
% %         spotprops(ii).fillglbound = any(db.isbound);
% %     
% %     catch ME
% %         % for now if this generates an error, NaN things out and continue.
% %         spotprops(ii).fillglfitdata = NaN;
% %         spotprops(ii).fillglr2 = -1;
% %         spotprops(ii).fillglsum = NaN;
% %         spotprops(ii).fillglbound = NaN;
% %     end
    
    
    try
        [temp,r2,db] = gausswfitlayers(subreg(:),xdata,[],wdata);
        spotprops(ii).subglfitdata = temp;
        spotprops(ii).subglr2 = r2;
        spotprops(ii).subglsum = sum(flatten(db.withoutbg));
        spotprops(ii).subglbound = any(db.isbound);
    
    catch ME
        % for now if this generates an error, NaN things out and continue.
        spotprops(ii).subglfitdata = NaN;
        spotprops(ii).subglr2 = -1;
        spotprops(ii).subglsum = NaN;
        spotprops(ii).subglbound = NaN;
    end
    
    
end

