function [afit,r2,debug] = gausswfitlayers(ydata,xdata,a0,wwid,options)

% fit a 2D Gaussian to the data in the ydata array
% to begin with, reshape everything within this function

% This weighted version allows the central pixels to be weighted more
% strongly than peripheral ones.  To do this we need to switch from
% lsqcurvefit to lsqnonlin..
% This means calculating the difference between the fit and the data,
% rather than just the fit.

if isvector(ydata)
    siz = sqrt(numel(ydata));
else
    siz = size(ydata,1);
    ydata = ydata(:);
end

if nargin<2 || isempty(xdata)
    xdata = adind2sub([siz,siz],1:(siz^2));
end

if nargin<4 || isempty(wwid)
    if nargin<3 || isempty(a0)
        wwid = siz/2;
    else
        % use our supplied sigma as a guide
        wwid = 2*a0(4);
    end
end

if numel(wwid)==1
    % generate a weighting mask then flatten it
    wdata = exp(-((xdata(:,1)-(siz+1)/2).^2 + (xdata(:,2)-(siz+1)/2).^2)/(2*wwid^2));
else
    wdata = wwid(:);
end

zdim = max(xdata(:,3));
cc = (max(xdata(:,1:2),[],1)+1)/2;

% in this version we need a separate amplitude and x- and y-centre for each z-slice
% allow a different background too?
if nargin<3 || isempty(a0)
    % Parameters:
    % 1 - sigma
    % 2nd - amplitude
    % 3rd - x-centre
    % 4th - y-centre
    % 5th - background
    
    % intially they can all be the same
    
    
    bg = admean(ydata(min(xdata,[],2)==1 | max(xdata,[],2)==siz));
    a0temp = [max(ydata) - bg, (siz+1)/2, (siz+1)/2, bg];
    
    
    temp = ones(zdim,1)*a0temp;
    
    a0 = [siz/4; temp(:)];
end

if nargin<5 || isempty(options)
% %     options = optimset('Algorithm','levenberg-marquardt','display','off');
    options = optimset('Algorithm','trust-region-reflective','display','off');
end

if nnz(isnan(ydata))>0
    xdata = xdata(~isnan(ydata),:);
    wdata = wdata(~isnan(ydata));
    ydata = ydata(~isnan(ydata));
end

a0 = a0(:);

fun = @(a) (a(1 + xdata(:,3)).*exp(-((xdata(:,1)-a(1+zdim+xdata(:,3))).^2 + ...
    (xdata(:,2)-a(1+2*zdim+xdata(:,3))).^2)/(2*a(1)^2))+a(1+3*zdim+xdata(:,3)) - ydata).*wdata;

% % fun = @(a) subfcn(a,xdata,ydata,wdata,zdim,cc);

% could impose our own upper and lower bounds (but still use levenberg
% marquardt) by penalizing bad parameter values..

lb = [0.4;zeros(zdim,1);(cc(1)-3)*ones(zdim,1);(cc(2)-3)*ones(zdim,1);ones(zdim,1)*min([0;ydata])];
ub = [4;2*(max(ydata)-min(ydata))*ones(zdim,1);(cc(1)+3)*ones(zdim,1);(cc(2)+3)*ones(zdim,1);ones(zdim,1)*max(ydata)];


try
%     [afit,resnorm] = lsqnonlin(fun,a0,[],[],options);
    [afit,resnorm] = lsqnonlin(fun,a0,lb,ub,options);
catch ME
    % keyboard
    % the most likely reason for a crash here is a bunch of NaNs messing
    % things up.
    % When that happens, just NaN everything out
    afit = NaN*ones(size(a0));
    
    % speed not really an issue here
    r2 = NaN;
    debug.withbg = NaN*ones(size(xdata,1),1);
    debug.withoutbg = NaN*ones(size(xdata,1),1);
    return
end

if nargout>1
    ytot = sum((ydata-mean(ydata)).^2);
    r2 = 1 - resnorm/ytot;
    if nargout>2
        % try returning the best fit including the background and without
        % it
        debug.withbg = afit(1 + xdata(:,3)).*exp(-((xdata(:,1)-afit(1+zdim+xdata(:,3))).^2 + ...
            (xdata(:,2)-afit(1+2*zdim+xdata(:,3))).^2)/(2*afit(1)^2))+afit(1+3*zdim+xdata(:,3));
        
        % because NaNs are removed at the start, don't bother reshaping
% %         debug.withbg = reshape(withbg,max(xdata,[],1));
        
        debug.withoutbg = afit(1 + xdata(:,3)).*exp(-((xdata(:,1)-afit(1+zdim+xdata(:,3))).^2 + ...
            (xdata(:,2)-afit(1+2*zdim+xdata(:,3))).^2)/(2*afit(1)^2));
% %         debug.withoutbg = reshape(withoutbg,max(xdata,[],1));
        
        % also return whether any of the fitting parameters are at the
        % upper or lower limits, indicating a bad (or non-spot) fit
        debug.isbound = afit==lb | afit==ub;

    end
end

% % function val = subfcn(a,xdata,ydata,wdata,zdim,cc)
% % 
% % val = (a(1 + xdata(:,3)).*exp(-((xdata(:,1)-a(1+zdim+xdata(:,3))).^2 + ...
% %     (xdata(:,2)-a(1+2*zdim+xdata(:,3))).^2)/(2*a(1)^2))+a(1+3*zdim+xdata(:,3)) - ydata).*wdata;
% % 
% % % also increase val if the centres are too far away
% % deltax = a(1+zdim+(1:zdim)) - cc(1);
% % deltay = a(1+2*zdim+(1:zdim))- cc(2);
% % 
% % d2 = deltax.^2 + deltay.^2;
% % 
% % 
% % 
% % val = val*4*sum(d2>100);

