function [trackprops,tracks] = measuretrackprops(cellprops,tracks,pixsize)

% make a structure with information about individual tracks.
% Things which we can measure for tracks and not individual cell frames:
%
% - Cell volume (since for individual frames this is quite noisy)
% - Cell footprint area, mean and variance
% - MSD
% - Non-Gaussianity of displacement distribution?
% - vector of distance to nearest cell in each frame of the track
% - vector of 2D displacements in each frame of the track

% we can also record in a track-by-track manner things like distance moved
% in the next frame, gene expression maximum/sum, frequency of gene
% expression, that will be useful to have in a separate form for further
% analysis.
% Having them in this form allows us to quickly perform the averaging over
% the whole dataset, or over individual tracks.

if nargin<3 || isempty(pixsize)
    pixsize = [0.24,0.24,0.33];
end

if nargin<2 || isempty(tracks)
    if ~sum(strcmpi(fieldnames(cellprops),'nextcells'))
        cellprops = performtracking3(cellprops);
    end
    tracks = findgoodtracks(cellprops);
end

[x,y] = props2tracks(cellprops,tracks,'xy');

vol = props2tracks(cellprops,tracks,'vol');
nucgfp = props2tracks(cellprops,tracks,'nucgfp');
area = props2tracks(cellprops,tracks,'footprint');

% find how many spot methods have been used
fnames = fieldnames(cellprops);
isspotprop = strfind(fnames,'spotprops');
count = 0;
for ii = 1:numel(fnames)
    if ~isempty(isspotprop{ii})
        count = count + 1;
        spotpropname{count} = fnames{ii};
    end
end

if ~sum(strcmpi(fnames,'cellid'))
    % assign cellids if not done already
    for ii = 1:numel(cellprops)
        cellprops(ii).cellid = ii;
    end
end
cellid = props2tracks(cellprops,tracks,'cellid');

if sum(strcmpi(fnames,'video'))
    vids = 1;
else
    vids = 0;
end

trackprops(size(tracks,2)).medvol = NaN;

frames = props2tracks(cellprops,tracks,'frame');

for ii = 1:size(tracks,2)
    % important housekeeping information is the video and the cellids
    if vids
        trackprops(ii).video = cellprops(tracks(1,ii)).video;
    end
    trackprops(ii).cellid = cellid(~isnan(cellid(:,ii)),ii);
    
    trackprops(ii).frames = frames(~isnan(frames(:,ii)),ii);
    trackprops(ii).tracklen = nnz(~isnan(x(:,ii)));
    
    trackprops(ii).medvol = admedian(vol(:,ii));
    trackprops(ii).meanvol = admean(vol(:,ii));
    trackprops(ii).stdvol = adstd(vol(:,ii));
    
    trackprops(ii).medarea = admedian(area(:,ii));
    trackprops(ii).meanarea = admean(area(:,ii));
    trackprops(ii).stdarea = adstd(area(:,ii));
    % change in footprint area?
    trackprops(ii).diffarea = diff(area(1:trackprops(ii).tracklen,ii));
    
    trackprops(ii).nucgfp = nucgfp(1:trackprops(ii).tracklen,ii);
    
    tmax = max(3,ceil(trackprops(ii).tracklen/2));
    for time = 2:tmax
        lag = time-1;
        dx = (x(time:lag:end,ii) - x(1:lag:(end-lag),ii));
        dy = (y(time:lag:end,ii) - y(1:lag:(end-lag),ii));
        use = ~isnan(dx);
        
        if time==2
            % record the frame-by-frame displacement
            trackprops(ii).framedisp = sqrt(dx(use).^2 + dy(use).^2);
            trackprops(ii).xdisp = dx(use);
            trackprops(ii).ydisp = dy(use);
            trackprops(ii).dispang = atan2(dy(use),dx(use));
        end
        
        n = numel(dx(use));
        
        trackprops(ii).MSD(time,1) = mean(dx(use).^2 + dy(use).^2);
        trackprops(ii).MSD(time,2) = sqrt(((sum(dx(use).^4) - n*(mean(dx(use).^2))^2)/((n-1)*n)) + ...
            ((sum(dy(use).^4) - n*(mean(dy(use).^2))^2)/((n-1)*n)));
        
    end
    
    for jj = 1:trackprops(ii).tracklen
        if ~(isnan(tracks(jj,ii)) || tracks(jj,ii)==0)
            try
            if ~isempty(cellprops(tracks(jj,ii)).nucxy)
                trackprops(ii).nucxyz(jj,:) = cellprops(tracks(jj,ii)).nucxy;
            else
                trackprops(ii).nucxyz(jj,1:3) = NaN;
            end
            catch ME
                rethrow(ME)
            end
        else
            trackprops(ii).nucxyz(jj,1:3) = NaN;
        end
    end
    
    trackprops(ii).rmsdisp = sqrt(trackprops(ii).MSD(2,1));
    trackprops(ii).rmsdisperror = 0.5*trackprops(ii).rmsdisp*trackprops(ii).MSD(2,2)/trackprops(ii).MSD(2,1);
    
    if sum(strcmpi(fnames,'d'))
        [d1,d2,d3,d4] = props2tracks(cellprops,tracks(tracks(:,ii)>0,ii),'d');
        [v1,v2,v3,v4] = props2tracks(cellprops,tracks(tracks(:,ii)>0,ii),'v');
        
        % check that the d's and v's are the right size
        if isempty(d4) || isempty(v4)
            d1 = NaN*ones(nnz(tracks(:,ii)),1);
            d2=d1; d3=d1; d4=d1; v1=d1; v2=d1; v3=d1; v4=d1;
        end
        
        trackprops(ii).aspr = sqrt(max(d1./d4,d4./d1));
        vmajor = repmat((d1>d4),[1 2]).*[v1,v2] + repmat((d1<=d4),[1 2]).*[v3,v4];
        trackprops(ii).longax = atan(vmajor(:,2)./vmajor(:,1));
    end
    
    
    % also record the expression intensity profile for each method
    
    for jj = 1:count
        methnum = spotpropname{jj}(end);
        for kk = 1:trackprops(ii).tracklen
%             if kk==73
%                 keyboard
%             end
            if (~isnan(tracks(kk,ii))) && (tracks(kk,ii)~=0) && (~isempty(cellprops(tracks(kk,ii)).(spotpropname{jj}))) && (~isempty(cellprops(tracks(kk,ii)).(spotpropname{jj})(1).maxspotint))
                % try to subtract off background
                
                trackprops(ii).(['maxspotint' methnum])(kk,1) = cellprops(tracks(kk,ii)).(spotpropname{jj})(1).maxspotint ...
                    - cellprops(tracks(kk,ii)).(spotpropname{jj})(1).bgint;
                
                trackprops(ii).(['spotint' methnum])(kk,1) = cellprops(tracks(kk,ii)).(spotpropname{jj})(1).spotint ...
                    - cellprops(tracks(kk,ii)).(spotpropname{jj})(1).bgint;
                
                trackprops(ii).(['corrsum' methnum])(kk,1) = cellprops(tracks(kk,ii)).(spotpropname{jj})(1).corrsum;
                try
                    trackprops(ii).(['filtresp' methnum])(kk,1) = cellprops(tracks(kk,ii)).(spotpropname{jj})(1).filtresp;
                end
                try
                trackprops(ii).(['spotxyz' methnum])(kk,:) = cellprops(tracks(kk,ii)).(spotpropname{jj})(1).xy;
                catch ME
                    rethrow(ME)
                end
                
                % also want to record the nuclear position of the spot

                if ~isempty(cellprops(tracks(kk,ii)).nucarea) && ~isempty(cellprops(tracks(kk,ii)).(spotpropname{jj})(1).nucedgedist)
%                     trackprops(ii).(['nucedgedist' methnum])(kk,1) = cellprops(tracks(kk,ii)).(spotpropname{jj})(1).nucedgedist;
                    % recalculate the nuclear position in microns using the
                    % supplied pixel size
                    ndim = size(cellprops(tracks(kk,ii)).nucedgepix,2);
                    nedge = size(cellprops(tracks(kk,ii)).nucedgepix,1);
                    spotpos = trackprops(ii).(['spotxyz' methnum])(kk,1:ndim);
                    
                    nucedgedist = distance(pixsize(1:ndim).*spotpos,...
                        repmat(pixsize(1:ndim),[nedge,1]).*cellprops(tracks(kk,ii)).nucedgepix(:,[2,1,3]));
                    
                    % find which edge pixel the spot is closest to, then
                    % based on whether this edge pixel is closer to the
                    % nuclear centre than the spot, tells us whether the
                    % spot is inside or outside the segmented nuclear
                    % region.
                    % if it's outside, change the sign of the distance.
                    mindist = min(nucedgedist);
                    edgeind = find(nucedgedist==mindist,1,'first'); % one only should be good enough
                    
                    spotvect = (spotpos - cellprops(tracks(kk,ii)).nucxy);
                    spotdir = spotvect/norm(spotvect);
                    edgeoffset = 0.5*sum(abs(spotdir).*pixsize);
                    
                    edgdist2 = sum((pixsize(1:ndim).*(cellprops(tracks(kk,ii)).nucedgepix(edgeind,[2,1,3]) - cellprops(tracks(kk,ii)).nucxy).^2));
                    spotdist2 = sum((pixsize(1:ndim).*spotvect).^2);
                    try
                        trackprops(ii).(['nucedgedist' methnum])(kk,1) = (mindist+edgeoffset).*(1-2*double(spotdist2>edgdist2));
                    catch ME
                        rethrow(ME)
                    end
                else
                    trackprops(ii).(['nucedgedist' methnum])(kk,1) = NaN;
                end

                
            else
                
                trackprops(ii).(['maxspotint' methnum])(kk,1) = NaN;
                trackprops(ii).(['nucedgedist' methnum])(kk,1) = NaN;
                trackprops(ii).(['spotint' methnum])(kk,1) = NaN;
                trackprops(ii).(['corrsum' methnum])(kk,1) = NaN;
                trackprops(ii).(['filtresp' methnum])(kk,1) = NaN;
                trackprops(ii).(['spotxyz' methnum])(kk,1:3) = NaN;
%                 if kk==73 && ii==1
%                     keyboard
%                 end
            end
        end
%         trackprops(ii).(['spotfreq' methnum]) = nnz(trackprops(ii).(['maxspotint' methnum]))/trackprops(ii).tracklen;
    end
    
    if sum(strcmpi(fnames,'nndist'))
        trackprops(ii).nndist = props2tracks(cellprops,tracks(1:trackprops(ii).tracklen,ii),'nndist');
        trackprops(ii).meannndist = admean(trackprops(ii).nndist);
        trackprops(ii).nnang = props2tracks(cellprops,tracks(1:trackprops(ii).tracklen,ii),'nnang');
        trackprops(ii).nnid = props2tracks(cellprops,tracks(1:trackprops(ii).tracklen,ii),'nnind');
    end
    
    if sum(strcmpi(fnames,'localdens'))
        % if tracks can contain zeros where a cell disappears, need to
        % either make a new field2val2 or do it elementwise
        for jj = 1:trackprops(ii).tracklen
            if tracks(jj,ii)>0
                trackprops(ii).ldens(jj,1:numel(cellprops(tracks(jj,ii)).localdens)) = cellprops(tracks(jj,ii)).localdens;
            else
                trackprops(ii).ldens(jj,:) = NaN;
            end
        end
        
    end
    
end

