function [matchfrac,D0] = munkresndim(prevcoords,nextcoords,cutoff,reps,noise,efunc,ifunc1,ifunc2)

% prevcoords and nextcoords are n-dimensional, but we can use the other
% input arguments to work out which are the spatial coordinates and which
% are the intensity dimensions (which have different cost functions)

% A side effect of adding extra intensity dimensions is that the typical
% true prev-to-next cost will go up, which means that the cost of objects
% disappearing is slightly less penalizing, so non-matches might propagate
% in from the edge slightly, unless a scaled efunc is supplied to account
% for this.

% intensity functions should be specified outside this function, a
% potentially useful one is:
% gcost = @(x) 0.2*cutoff^2 * funcutoff(x,@(x)exp(-x/120),[50,600])
% where the three arbitrary numbers might need scaling depending on the
% typical noise in the intensity measurement.

totdim = size(prevcoords,2);
if nargin<7 || (isempty(ifunc1) && isempty(ifunc2)) || ...
        ((~isa(ifunc1,'function_handle') && isnan(ifunc1)) && (~isa(ifunc2,'function_handle') && isnan(ifunc2)))
    % no intensity function, so all dimensions are spatial
    dinds = 1:totdim;
    i1inds = [];
    i2inds = [];
elseif nargin==7 || isempty(ifunc2)
    % one intensity function, last coordinate dimension is intensity
    dinds = 1:(totdim-1);
    i1inds = totdim;
    i2inds = [];
else
    % two intensity functions, last two coordinate dimensions are
    % intensity.
    dinds = 1:(totdim-2);
    i1inds = totdim-1;
    i2inds = totdim;
end

if nargin<6 || isempty(efunc)
    % efunc = @(x) x.^2;
    efunc = @(x) (0.1*cutoff + 0.9*x).^2;
end

if nargin<5 || isempty(noise)
    noise = 0.02;
end

if nargin<4 || isempty(reps)
    reps = 200;
end

pobj = size(prevcoords,1);
nobj = size(nextcoords,1);
% options for the edge dependence of the disappearance cost are:
% - linear up to the cutoff: cutoff*dist
% - quadratic up to the cutoff: dist^2
% Try the quadratic to see if it works

try
D0 = constructcostmat(prevcoords(:,dinds),nextcoords(:,dinds),cutoff,[1,1,;512,512],efunc);
catch ME
    rethrow(ME)
end
if ~isempty(i1inds)
    D1 = constructcostmat(prevcoords(:,i1inds),nextcoords(:,i1inds),cutoff,[],[],ifunc1,1);
    D0(1:pobj,1:nobj) = D0(1:pobj,1:nobj) + D1;
end
if ~isempty(i2inds)
    D2 = constructcostmat(prevcoords(:,i2inds),nextcoords(:,i2inds),cutoff,[],[],ifunc2,1);
    D0(1:pobj,1:nobj) = D0(1:pobj,1:nobj) + D2;
end

tgind = zeros(reps,size(D0,1));
tcost = zeros(reps,1);

parfor ii = 1:reps
% %     matcharray(ii,:) = hungarianlinker(xy1,xy2 + noise*randn(size(xy2)),cutoff);
    D = D0;
    D(1:pobj,1:nobj) = D(1:pobj,1:nobj).*(1 + noise*randn(pobj,nobj));
    [tgind(ii,:),tcost(ii)] = munkres(D);
    % probably the best way of parallelizing this is by recording the tgind
    % values, because this is the same every time.
    
    
end

matchfrac = zeros(pobj,nobj);

for ii = 1:reps
    inds = find(tgind(ii,:)>0 & tgind(ii,:)<=nobj & (1:size(tgind,2))<=pobj)';
    linind = adsub2ind([pobj,nobj],[inds(:),tgind(ii,inds)']);
    matchfrac(linind) = matchfrac(linind) + 1/reps; % this is where we would weight using the total cost..
end


% % matchinds = find(tgind>0 & tgind<=nobj & (1:numel(tgind))<=pobj)';

% % matches = mode(matcharray,1);
% % 
% % frac = mean(matcharray==repmat(matches,[reps,1]),1);

