function cellprops = performhginttracking(cellprops,usexy,cutoff,conf,noise,reps,gfunc,rfunc)

% version of the cell tracking which uses the Hungarian algorithm to find
% the absolute best SSD for each frame-frame.
% We might need to do some checking afterwards, or if it's fast enough, we
% could even add some random noise and see how often the matches are
% disrupted..

% keep the same structure as for the original tracking

% IMPORTANT!
% At the moment we haven't altered the cost of unmatched objects from the
% default given in the hungarianlinker algorithm.  To do this, we should
% add an extra column for unmatched to the diffmat2, and pass this directly
% to the hungarian mex file.

% What does the cutoff actually do for the Hungarian tracker?
% Presumably it can be made quite large..?
if nargin<7 || isempty(gfunc)
    gfunc = @(x) 0.2*60^2 * funcutoff(x,@(x)exp(-x/120),[50,300]);
end
if nargin<8 || isempty(rfunc)
    rfunc = @(x) 0.2*60^2 * funcutoff(x,@(x)exp(-x/90),[40,200]);
end

if nargin<4 || isempty(conf)
    conf = 0; % by default, accept the lowest SSD configuration.
end

% with MC noise, it's wise to increase the cutoff to perhaps twice the
% typical displacement. But this can have the effect that distant objects
% get matched up rather than being labelled as new tracks.
if nargin<3 || isempty(cutoff)
    cutoff = 50;
end

% If we specify a non-zero noise, then add some noise to the target
% coordinates and repeat to see how robust the match is.
% The default value for the noise as a fraction of the cutoff takes account
% of the fact that we should have increased the cutoff for the MC method.
if nargin<5 || isempty(noise)
    noise = 0.125;
end

if nargin<6 || isempty(reps)
    reps = 400; % maybe this should be linked to the number of objects?
end

if noise==0;
    reps = 1; % no point repeating, will be the same every time!
end

fnames = fieldnames(cellprops);

usefield = firstmatch(fnames,{'xy','nucxyz','Centroid'});

% % if ~sum(strcmpi(fnames,'xy'))
% %     usefield = 'nucxyz';
% % else
% %     usefield = 'xy';
% % end

realxy = field2val2(cellprops,usefield);
realxy = realxy(:,1:2);
% used for edge distances during tracking

if nargin<2 || isempty(usexy)
    usexy = realxy;
end

if ischar(usexy)
    usexy = field2val2(cellprops,usexy);
    usexy = usexy(:,1:2);
end
% at the moment, just use xy distances
% there is the possibility of using other 'displacements' like volume or
% intensity only when there is ambiguity in the xy distance, then the units
% of the distance aren't important.


% If tracking has already been performed, we first need to remove the
% nextcells and prevcells labels
fnames = fieldnames(cellprops);
if sum(strcmpi(fnames,'nextcells'))
    cellprops = rmfield(cellprops,{'nextcells','prevcells','nextdists','prevdists'});
end

% do the intensity weighting here and add to usexy, rather than having to
% sort out the coordinates later
% if gfpw>0
%     ngfp = field2val2(cellprops,intstr);
% else
%     ngfp = [];
% end
% if rfpw>0
%     nrfp = field2val2(cellprops,'nucsumint')./field2val2(cellprops,'nucarea');
% else
%     nrfp = [];
% end
% 
% usexy = [usexy,gfpw*ngfp,rfpw*nrfp];

% if isempty(fid)
%     fid = fopen('trackinglog.txt','wt');
%     baselevel = 1;
% else
%     baselevel = 0;
% end

frame = field2val(cellprops,'frame');
% find number of frames
nframes = max(frame);


% catch ME
%     rethrow(ME)
% end
cellprops(1).prevcells = [];
cellprops(1).prevdists = [];
cellprops(1).nextcells = [];
cellprops(1).nextdists = [];

% label each cell with a number, to uniquely identify it when we're looking
% at different subsets of the whole dataset

% ONLY DO THIS IF IT'S NOT THERE ALREADY!
% this will need redoing in the case where individual videos have been
% tracked, and then merged, because there will be duplicate cellids.
% REMEMBER TO REMOVE CELLID MANUALLY FIRST IN THIS CASE!!

if ~sum(strcmpi(fnames,'cellid'))
    for ii = 1:numel(cellprops)
        cellprops(ii).cellid = ii;
    end
else
    % check for non-unique cellid values
    cellid = field2val2(cellprops,'cellid');
    if numel(unique(cellid))<numel(cellid)
        cellprops = rmfield(cellprops,'cellid');
        for ii = 1:numel(cellprops)
            cellprops(ii).cellid = ii;
        end
    end
end


if sum(strcmpi(fnames,'video')) && sum(abs(diff([cellprops.video])))
    % call performtracking3 recursively for each video
    video = field2val2(cellprops,'video');
    cellprops(1).nextcells = [];
    cellprops(1).prevcells = [];
    cellprops(1).nextdists = [];
    cellprops(1).prevdists = [];
    
    % see if we've assigned different cutoffs to the different videos, and
    % if not, replicate the cutoff
    if isscalar(cutoff)
        cutoff = cutoff*ones(1,max(video));
    end
    
    for ii = 1:max(video) % might be gaps in the video numbering
        if nnz(video==ii)
%             fprintf(fid,'Video %d \n',ii);
            tcellprops = performhginttracking(cellprops(video==ii),usexy(video==ii,:),cutoff(ii),conf,noise,reps,gfunc,rfunc);
            try
            outcellprops(video==ii) = tcellprops; % this line could be merged with above?
            catch ME
                rethrow(ME)
            end
        end
    end
    cellprops = outcellprops;
else
    wb = waitbar(0,'initialising');
    % Initialise prevcoords for first iteration of loop
    previnds = find(frame==1);
    % try
    % %     prevcoords = field2val2(cellprops(previnds),'xy');
    prevcoords = usexy(previnds,:);
    % %     prevcoords = prevcoords(:,1:2); % don't include z
    %     if rfpw>0
    %         prevvol = field2val2(cellprops(previnds),'vol');
    %         prevcoords = [prevcoords, rfpw*prevvol];
    %     end
    %     if gfpw>0
    %         prevint = field2val2(cellprops(previnds),intstr);
    %         prevcoords = [prevcoords, gfpw*prevint];
    %     end
    
    rprev = realxy(previnds,:);
    
    for ii = 2:nframes
        % construct xy vector for next frame
        nextinds = find(frame==ii);
% %         nextcoords = field2val2(cellprops(nextinds),'xy');
        nextcoords = usexy(nextinds,:);
% %         nextcoords = nextcoords(:,1:2); % no need to include z
        
        rnext = realxy(nextinds,:);
        
%         
%         if rfpw>0
%             nextvol = field2val2(cellprops(nextinds),'vol');
%             nextcoords = [nextcoords, rfpw*nextvol];
%         end
%         if gfpw>0
%             nextint = field2val2(cellprops(nextinds),intstr);
%             nextcoords = [nextcoords,gfpw*nextint];
%         end
        
        % munkresmc will construct the distance matrix
        % with the arbitrary cost functions, the GFP, and H2B intensities
        % can be fed in without overly penalizing noisy measurements
        [matchfrac,D0] = munkresndim(prevcoords,nextcoords,cutoff,reps,noise,[],gfunc,rfunc);
        % matchinds = findn(matchfrac>conf); % this will be unique as long as conf>0.5, otherwise we have to find the max
        maxval = max(matchfrac,[],2);
        matchinds = findn(matchfrac>conf & matchfrac==repmat(maxval,[1,size(matchfrac,2)]));

        % need to allow for the fact that either prevcoords or nextcoords could
        % be empty.
        
        % would be better to record all max matches, and separately their
        % confidence value, so we can choose later what value of confidence
        % is the right one to use.
        
        waitbar((ii-1)/(nframes-1),wb,'Tracking..')
        
        for jj = 1:size(matchinds,1)
            cellprops(previnds(matchinds(jj,1))).nextcells = [cellprops(previnds(matchinds(jj,1))).nextcells, cellprops(nextinds(matchinds(jj,2))).cellid];
            cellprops(nextinds(matchinds(jj,2))).prevcells = [cellprops(nextinds(matchinds(jj,2))).prevcells, cellprops(previnds(matchinds(jj,1))).cellid];
            cellprops(previnds(matchinds(jj,1))).nextdists = [cellprops(previnds(matchinds(jj,1))).nextdists, sqrt(D0(matchinds(jj,1),matchinds(jj,2)))];
            cellprops(nextinds(matchinds(jj,2))).prevdists = [cellprops(nextinds(matchinds(jj,2))).prevdists, sqrt(D0(matchinds(jj,1),matchinds(jj,2)))];
        end

        % update previnds and prevcoords ready for the next iteration
        previnds = nextinds;
        prevcoords = nextcoords;
        rprev = rnext;
    end
    close(wb)
end

% if baselevel
%     fclose(fid);
% end
