function cellprops = performtracking3(cellprops,cutoff,intweight,volweight)

% track cell objects from frame to frame, by adding to cellprops the
% indices of the objects which match in the previous and next frames (along
% with their xy and hybrid distances in order to choose the best?)

% from there, we'll need another function to sort and collect the matches,
% or view the possibilities.  To decide between cell divisions or overlaps,
% it would be good to track the predicted/mean/probable number of cells
% contained in each object.
% Also, objects near the edge should be interpreted differently to objects
% in the middle of the field of view, since edge objects have a higher
% probability of disappearing.
% Come up with the rules for this once we know it's needed.

% we want symmetry in the track indices, so that if cell a appears in the
% next field of cell b, cell b must appear in the prev field of cell a.

% For use on merged datasets of more than one video, we need to use the
% 'video' field as well as the frame to match cells

if nargin<4 || isempty(volweight)
    volweight = 0;
end
if nargin<3 || isempty(intweight)
    intweight = 0;
end

if nargin<2 || isempty(cutoff)
    cutoff = [20 40 80];
end
% if numel(cutoff)==1
%     cutoff = [0.5 1 2]*cutoff; % will be slower, but leave for now
% end

cutoff = sort(cutoff,'ascend');


% at the moment, just use xy distances
% there is the possibility of using other 'displacements' like volume or
% intensity only when there is ambiguity in the xy distance, then the units
% of the distance aren't important.


% If tracking has already been performed, we first need to remove the
% nextcells and prevcells labels
fnames = fieldnames(cellprops);
if sum(strcmpi(fnames,'nextcells'))
    cellprops = rmfield(cellprops,{'nextcells','prevcells','nextdists','prevdists'});
end

% if there are multiple videos in one dataset, need to add this.

frame = field2val(cellprops,'frame');
% find number of frames
nframes = max(frame);

% Initialise prevcoords for first iteration of loop
previnds = find(frame==1);
try
    prevcoords = field2val2(cellprops(previnds),'xy');
    prevcoords = prevcoords(:,1:2); % don't include z
    prevvol = field2val2(cellprops(previnds),'vol');
    if volweight>0
        prevcoords = [prevcoords, volweight*prevvol];
    end
    if intweight>0
        prevint = field2val2(cellprops(previnds),'gfpint');
        prevcoords = [prevcoords, intweight*prevint];
    end
catch ME
    rethrow(ME)
end
cellprops(1).prevcells = [];
cellprops(1).prevdists = [];
cellprops(1).nextcells = [];
cellprops(1).nextdists = [];

% label each cell with a number, to uniquely identify it when we're looking
% at different subsets of the whole dataset

% ONLY DO THIS IF IT'S NOT THERE ALREADY!
if ~sum(strcmpi(fnames,'cellid'))
    for ii = 1:numel(cellprops)
        cellprops(ii).cellid = ii;
    end
end


if sum(strcmpi(fnames,'video')) && sum(abs(diff([cellprops.video])))
    % call performtracking3 recursively for each video
    video = [cellprops.video];
    cellprops(1).nextcells = [];
    cellprops(1).prevcells = [];
    cellprops(1).nextdists = [];
    cellprops(1).prevdists = [];
    
    for ii = 1:max(video) % might be gaps in the video numbering
        if nnz(video==ii)
            tcellprops = performtracking3(cellprops(video==ii),cutoff,intweight,volweight);
            try
            outcellprops(video==ii) = tcellprops; % this line could be merged with above?
            catch ME
                rethrow(ME)
            end
        end
    end
    cellprops = outcellprops;
else

    for ii = 2:nframes
        % construct xy vector for next frame
        nextinds = find(frame==ii);
        nextcoords = field2val2(cellprops(nextinds),'xy');
        nextcoords = nextcoords(:,1:2); % no need to include z
        
        nextvol = field2val2(cellprops(nextinds),'vol');
        if volweight>0
            nextcoords = [nextcoords, volweight*nextvol];
        end
        if intweight>0
            nextint = field2val2(cellprops(nextinds),'gfpint');
            nextcoords = [nextcoords,intweight*nextint];
        end
        
        % need to allow for the fact that either prevcoords or nextcoords could
        % be empty.

        % create distance matrix and find the elements smaller than cutoff
        % if volweight or intweight are non-zero, add these to the
        % coordinates.
        
        prevmat = repmat(permute(prevcoords,[1 3 2]),[1,size(nextcoords,1),1]);
        nextmat = repmat(permute(nextcoords,[3,1,2]),[size(prevcoords,1),1,1]);

        diffmat2 = sum((nextmat - prevmat).^2,3);

        freenext = true(numel(nextinds),1);
        freeprev = true(numel(previnds),1);

        for kk = 1:numel(cutoff)
            % Repeat the tracking for each cutoff value, starting at the
            % smallest
            remprevinds = find(freeprev);
            remnextinds = find(freenext);

            remprev = previnds(freeprev);
            remnext = nextinds(freenext);
            remdiffmat2 = diffmat2(freeprev,freenext);

            matchinds = findn(remdiffmat2<(cutoff(kk)^2));


            for jj = 1:size(matchinds,1)
                cellprops(remprev(matchinds(jj,1))).nextcells = [cellprops(remprev(matchinds(jj,1))).nextcells, cellprops(remnext(matchinds(jj,2))).cellid];
                cellprops(remnext(matchinds(jj,2))).prevcells = [cellprops(remnext(matchinds(jj,2))).prevcells, cellprops(remprev(matchinds(jj,1))).cellid];
                cellprops(remprev(matchinds(jj,1))).nextdists = [cellprops(remprev(matchinds(jj,1))).nextdists, sqrt(remdiffmat2(matchinds(jj,1),matchinds(jj,2)))];
                cellprops(remnext(matchinds(jj,2))).prevdists = [cellprops(remnext(matchinds(jj,2))).prevdists, sqrt(remdiffmat2(matchinds(jj,1),matchinds(jj,2)))];

                freenext(remnextinds(matchinds(jj,2))) = 0;
                freeprev(remprevinds(matchinds(jj,1))) = 0;
            end


        end

        % update previnds and prevcoords ready for the next iteration
        previnds = nextinds;
        prevcoords = nextcoords;
    end
end
