function cellprops = performtracking4(cellprops,cutoff,intweight,volweight,conf,intstr)

% track cell objects from frame to frame, by adding to cellprops the
% indices of the objects which match in the previous and next frames (along
% with their xy and hybrid distances in order to choose the best?)

% from there, we'll need another function to sort and collect the matches,
% or view the possibilities.  To decide between cell divisions or overlaps,
% it would be good to track the predicted/mean/probable number of cells
% contained in each object.
% Also, objects near the edge should be interpreted differently to objects
% in the middle of the field of view, since edge objects have a higher
% probability of disappearing.
% Come up with the rules for this once we know it's needed.

% we want symmetry in the track indices, so that if cell a appears in the
% next field of cell b, cell b must appear in the prev field of cell a.

% For use on merged datasets of more than one video, we need to use the
% 'video' field as well as the frame to match cells

if nargin<6 || isempty(intstr)
    intstr = 'gfpint';
end

if nargin<5 || isempty(conf)
    conf = 0.7;
end

if nargin<4 || isempty(volweight)
    volweight = 0;
end
if nargin<3 || isempty(intweight)
    intweight = 0;
end

if nargin<2 || isempty(cutoff)
    cutoff = 80;
end

cutoff = sort(cutoff,'ascend');

% at the moment, just use xy distances
% there is the possibility of using other 'displacements' like volume or
% intensity only when there is ambiguity in the xy distance, then the units
% of the distance aren't important.


% If tracking has already been performed, we first need to remove the
% nextcells and prevcells labels
fnames = fieldnames(cellprops);
if sum(strcmpi(fnames,'nextcells'))
    cellprops = rmfield(cellprops,{'nextcells','prevcells','nextdists','prevdists'});
end

% if there are multiple videos in one dataset, need to add this.

frame = field2val(cellprops,'frame');
% find number of frames
nframes = max(frame);

% Initialise prevcoords for first iteration of loop
previnds = find(frame==1);
try
    prevcoords = field2val2(cellprops(previnds),'xy');
    prevcoords = prevcoords(:,1:2); % don't include z
    prevvol = field2val2(cellprops(previnds),'vol');
    if volweight>0
        prevcoords = [prevcoords, volweight*prevvol];
    end
    if intweight>0
        prevint = field2val2(cellprops(previnds),intstr);
        prevcoords = [prevcoords, intweight*prevint];
    end
catch ME
    rethrow(ME)
end
cellprops(1).prevcells = [];
cellprops(1).prevdists = [];
cellprops(1).nextcells = [];
cellprops(1).nextdists = [];

% label each cell with a number, to uniquely identify it when we're looking
% at different subsets of the whole dataset

% ONLY DO THIS IF IT'S NOT THERE ALREADY!
% this will need redoing in the case where individual videos have been
% tracked, and then merged, because there will be duplicate cellids.
% REMEMBER TO REMOVE CELLID MANUALLY FIRST IN THIS CASE!!

if ~sum(strcmpi(fnames,'cellid'))
    for ii = 1:numel(cellprops)
        cellprops(ii).cellid = ii;
    end
else
    % check for non-unique cellid values
    cellid = field2val2(cellprops,'cellid');
    if numel(unique(cellid))<numel(cellid)
        cellprops = rmfield(cellprops,'cellid');
        for ii = 1:numel(cellprops)
            cellprops(ii).cellid = ii;
        end
    end
end


if sum(strcmpi(fnames,'video')) && sum(abs(diff([cellprops.video])))
    % call performtracking3 recursively for each video
    video = [cellprops.video];
    cellprops(1).nextcells = [];
    cellprops(1).prevcells = [];
    cellprops(1).nextdists = [];
    cellprops(1).prevdists = [];
    
    for ii = 1:max(video) % might be gaps in the video numbering
        if nnz(video==ii)
            tcellprops = performtracking4(cellprops(video==ii),cutoff,intweight,volweight,conf,intstr);
            try
            outcellprops(video==ii) = tcellprops; % this line could be merged with above?
            catch ME
                rethrow(ME)
            end
        end
    end
    cellprops = outcellprops;
else
    wb = waitbar(0,'initialising');
    for ii = 2:nframes
        % construct xy vector for next frame
        nextinds = find(frame==ii);
        nextcoords = field2val2(cellprops(nextinds),'xy');
        nextcoords = nextcoords(:,1:2); % no need to include z
        
        nextvol = field2val2(cellprops(nextinds),'vol');
        if volweight>0
            nextcoords = [nextcoords, volweight*nextvol];
        end
        if intweight>0
            nextint = field2val2(cellprops(nextinds),intstr);
            nextcoords = [nextcoords,intweight*nextint];
        end
        
        % need to allow for the fact that either prevcoords or nextcoords could
        % be empty.
        prevmat = repmat(permute(prevcoords,[1 3 2]),[1,size(nextcoords,1),1]);
        nextmat = repmat(permute(nextcoords,[3,1,2]),[size(prevcoords,1),1,1]);

        diffmat2 = sum((nextmat - prevmat).^2,3);
        
        matchinds = dendritictrack(prevcoords,nextcoords,cutoff,conf);
        
        waitbar((ii-1)/(nframes-1),wb,'Tracking..')
        
        bad = false(size(matchinds,1),1);
        for jj = 1:size(matchinds,1)
            if diffmat2(matchinds(jj,1),matchinds(jj,2))>(cutoff^2)
                bad(jj) = true;
            end
        end
        matchinds(bad,:) = [];
        

        for jj = 1:size(matchinds,1)
            cellprops(previnds(matchinds(jj,1))).nextcells = [cellprops(previnds(matchinds(jj,1))).nextcells, cellprops(nextinds(matchinds(jj,2))).cellid];
            cellprops(nextinds(matchinds(jj,2))).prevcells = [cellprops(nextinds(matchinds(jj,2))).prevcells, cellprops(previnds(matchinds(jj,1))).cellid];
            cellprops(previnds(matchinds(jj,1))).nextdists = [cellprops(previnds(matchinds(jj,1))).nextdists, sqrt(diffmat2(matchinds(jj,1),matchinds(jj,2)))];
            cellprops(nextinds(matchinds(jj,2))).prevdists = [cellprops(nextinds(matchinds(jj,2))).prevdists, sqrt(diffmat2(matchinds(jj,1),matchinds(jj,2)))];
        end

        % update previnds and prevcoords ready for the next iteration
        previnds = nextinds;
        prevcoords = nextcoords;
    end
    close(wb)
end
